# include "inet.h"

# define DEBUG

typedef struct{

  uint32_t        pid        : 16;
  uint32_t        msg_type   : 16;
  uint32_t        request_no     ;
  
}HEADER_CLIENTD_PACKET;

# define CLIENTD_ADDRESS "::1"

extern const struct in6_addr in6addr_any;
HEADER_CLIENTD_PACKET        header;

main(int argc, char *argv[]){
  
  int                 sockfd, lib_sockfd, n, lib_socklen;
  struct sockaddr_in6 lib_sockaddr, clientd_sockaddr;
  char                recv_msg[MAX_MSG_SIZE + 1], send_msg[MAX_MSG_SIZE];
  char                name[128];
  uint32_t            *temp;

  sockfd = socket(AF_INET6, SOCK_STREAM, 0);

  if(sockfd < 0)
    err_dump("Error in opening datagram socket! Exiting...");


  bzero((char *) &clientd_sockaddr, sizeof(clientd_sockaddr));

#ifdef SIN6_LEN
  clientd_sockaddr.sin6_len      = sizeof(clientd_sockaddr);      /* 4.3 BSD doesn't have this field      */
#endif

  clientd_sockaddr.sin6_family   = AF_INET6;                      /* Family as required for Ipv6          */
  clientd_sockaddr.sin6_port     = htons(CLIENTD_PORT);           /* CLIENT Daemon listens on CLIENTD_PORT*/
  clientd_sockaddr.sin6_flowinfo = htonl(0);                      /* Flow & Traffic Class are unimportant */
  clientd_sockaddr.sin6_addr     = in6addr_any;                   /* CLIENTD_ADDRESS                      */
  clientd_sockaddr.sin6_scope_id = 0;                             /* Not really required.                 */

  if((n = bind(sockfd, (struct sockaddr *) &clientd_sockaddr, 
	  sizeof(clientd_sockaddr))) < 0){
    printf("bind: %d\n",n);
    printf("address = %s\n",inet_ntop(AF_INET6, &clientd_sockaddr.sin6_addr, name, 127));
    err_dump("Error while binding address! Exiting...");
  }

  listen(sockfd, 5);

  while(1){

    lib_socklen  = sizeof(lib_sockaddr);
    lib_sockfd   = accept(sockfd, (struct sockaddr *) &lib_sockaddr, &lib_socklen);
    if(lib_sockfd < 0)
      err_dump("Error in accept.Exiting...");

    n            = recv(lib_sockfd, recv_msg, MAX_MSG_SIZE, 0);
    if(n < 0)
      err_dump("Error in recv! Exiting...");

    memcpy(&header, recv_msg, sizeof(header));
    temp              = (uint32_t *) &header;
    *temp             = ntohl(*temp);
    header.request_no = ntohl(header.request_no);
    
#ifdef DEBUG
    printf("Received: Appl to client daemon: lib_addr = %s\n",
	   inet_ntop(AF_INET6, &lib_sockaddr.sin6_addr, name, 127));
    printf("msg_type   = %u\n", header.msg_type);
    printf("pid        = %u\n", header.pid);
    printf("request_no = %u\n", header.request_no);
#endif

    header.msg_type = 1;

#ifdef DEBUG
    printf("\n");
    printf("Sending: Client daemon to appl: lib_addr = %s\n",
	   inet_ntop(AF_INET6, &lib_sockaddr.sin6_addr, name, 127));
    printf("msg_type   = %u\n", header.msg_type);
    printf("pid        = %u\n", header.pid);
    printf("request_no = %u\n", header.request_no);
#endif

    temp              = (uint32_t *) &header;
    *temp             = htonl(*temp);
    header.request_no = htonl(header.request_no);
    n = sizeof(header);
    memcpy(send_msg, &header, n);
    if(send(lib_sockfd, send_msg, n, 0) != n)
      /*          \___________________________________
       *                                              \
       * For the moment let the server send using same port as it was listening on. 
       * You can change the port no. later for sending.
       */
      err_dump("Error in sendto! Exiting...");

  }
}
