/* Implemented by Dhawal Kumar */
# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>

# include "defines.h"

# include "../lib/adv.h"
# include "../lib/err_dump.h"
# include "../lib/error.h"
# include "../lib/solicit_params.h"
# include "../lib/header.h"

# include "../common/dhcp_parameters.h"
# include "../common/extensions.h"
# include "../common/dhcp_messages.h"
# include "../common/get_set_bits_utility.h"
# include "../common/read_write_data.h"
# include "../common/print_packet.h"
# include "../common/type_len_structure.h"



# include "appl_msg_data.h"
# include "interface_structures.h"
# include "list.h"

# include "globals.h"
# include "iaddr_if_list_ops.h"
# include "find.h"
# include "routines_for_solicit.h"
# include "send_error.h"
# include "send_msg.h"
# include "retransmit_msg.h"

# include "get_main_interface.h"
# include "discover_interfaces.h"
# include "create_solicit.h"
# include "handle_init.h"
# include "send_adv_to_appl.h"
# include "forward_reply_to_appl.h"
# include "handle_reply_from_server.h"
# include "handle_request_from_appl.h"
# include "try_cached_adv.h"
# include "retransmit_timedout_messages.h"
/*
 * Note: Important.......
 * (1) talks means both listens and replies.
 * (2) CLD currently listens on in6addr_any
 * (3) Applications must send explicit close to clientd and not use close() function.
 *     Why? Consider the following: An appl which has opened a socket descriptor say 5
 *     closes it. At the time the first appl closes connection another appl opens a descriptor
 *     which can be 5 if it the smallest available. Now, this appl sends init to CLD.
 *     If the two messages viz. (1) close() by 1'st and (2) connect() by 2'nd are 
 *     sent to CLD in reverse order i.e. (2) first, then at the time of arrival of close
 *     irrespective of what happens at the time of connect(), 5 will be unset from fd_set.
 *     This is not what one would expect. Note you have to design and implement a new message
 *     for this purpose. For the time being I am using ^C to kill the application and thus
 *     close connection. You can do this by sending zero sized message as well.
 */

main(int argc, char *argv[]){
  
  FD_INFO *temp_addr;
  TIME     temp_time;
  int      temp_int;

  fd_accept_list_head = NULL;
  fd_init_list_head   = NULL;
  in_list_head        = NULL;
  recv_msg = (char *) malloc(MAX_DHCP_MSG + APPL_DATA_OVERHEAD);
  send_msg = (char *) malloc(MAX_DHCP_MSG + APPL_DATA_OVERHEAD);
  FD_ZERO(&f);
  FD_ZERO(&temp_f);

  discover_interfaces();
  /* Imp:read lease file to check for consistency and update "obtained via DHCP" */

  /* Opening tcp_socket for listening to applications.                           */
  tcp_sockfd = socket(AF_INET6, SOCK_STREAM, 0);

  if(tcp_sockfd < 0)
    err_dump("Error in opening stream socket! Exiting...");
  
  /* Converting multicast addr "FF02::1:2" to binary for later use.              */
  n = inet_pton(AF_INET6, ALL_DHCP_AGENTS_ADDR,
		& server_multicast_addr);
  
  if(n <= 0){
    if(n < 0)
      err_dump("Unknown AF(first) argument. Exiting...");
    else
      err_dump("Invalid address string. Exiting...");
  }

  bzero((char *) &server_sockaddr, sizeof(server_sockaddr));

#ifdef SIN6_LEN
  server_sockaddr.sin6_len      = sizeof(server_sockaddr);
#endif

  server_sockaddr.sin6_family   = AF_INET6;
  server_sockaddr.sin6_port     = htons(AGENT_PORT);
  server_sockaddr.sin6_flowinfo = htonl(0);
  server_sockaddr.sin6_scope_id = 0;

  bzero((char *) &clientd_sockaddr, sizeof(clientd_sockaddr));

#ifdef SIN6_LEN
  clientd_sockaddr.sin6_len      = sizeof(clientd_sockaddr);
#endif

  clientd_sockaddr.sin6_family   = AF_INET6;
  clientd_sockaddr.sin6_port     = htons(CLD_TCP_PORT);
  clientd_sockaddr.sin6_flowinfo = htonl(0);
  clientd_sockaddr.sin6_addr     = in6addr_any;
  clientd_sockaddr.sin6_scope_id = 0;

  if((n = bind(tcp_sockfd, (struct sockaddr *) &clientd_sockaddr, 
	       sizeof(clientd_sockaddr))) < 0){
    printf("bind: %d\n",n);
    printf("address = %s\n",inet_ntop(AF_INET6, &clientd_sockaddr.sin6_addr, name, 127));
    err_dump("Error while binding address! Exiting...");
  }

  FD_SET(tcp_sockfd, &f);
  temp_f = f;
  /* 
   * Till now UDP sockets for listening and tcp socket
   * also for listening have been added to fd_set f.
   */
  nfds = (nfds > tcp_sockfd) ? nfds : tcp_sockfd;
  
  listen(tcp_sockfd, 5);

  SET_TIMEOUT(timeout, SLEEP_TIME * 100);


  while(1){
    loop++;
    n = select(nfds + 1, &f, (fd_set *)0, (fd_set *)0, &timeout);
    if(gettimeofday(&current_time, NULL) == -1)
      err_dump("Error in gettimeofday. Exiting...");

    if(n < 0)
      err_dump("Error in select\n\a");
    
    if(n == 0){
      f         = temp_f;
      temp_time = retransmit_timedout_messages();
    /* If you want the client daemon to delete expired leases do it here */
      if(temp_time.tv_sec == 0 && temp_time.tv_usec == 0){
	SET_TIMEOUT(timeout, SLEEP_TIME * 100);
      }
      else{

	if((temp_int = DIFF_IN_MICRO(temp_time, current_time)) > 0){
	  SET_TIMEOUT(timeout, temp_int);

	}
	else{
	  SET_TIMEOUT(timeout, 0);
	}
      }
      continue;
    }
    /* Now n > 0 */
    
    if(FD_ISSET(tcp_sockfd, &f)){
      /* A connect request by an application via library has arrived */
#ifdef DEBUG
      printf("Connect request by an appl arrived\n");
#endif

      FD_CLR(tcp_sockfd, &f);
      temp_accept     = (FD_ACCEPT_LIST *)malloc(sizeof(FD_ACCEPT_LIST));
      lib_socklen     = sizeof(lib_sockaddr);
      temp_accept->fd = accept(tcp_sockfd, (struct sockaddr *) &lib_sockaddr, &lib_socklen);

#ifdef DEBUG
      printf("Accept yields a new socket descriptor %d\n", temp_accept->fd);
#endif

      if(temp_accept->fd < 0)
	err_dump("Accept error. Exiting...");
      /*
       * Imp: Here you can check if request for connect came by an application on
       * same machine by checking lib_sockaddr. If not free(temp_accept) and don't
       * insert or set fd_set.
       */
      insert(temp_accept, &fd_accept_list_head);
      FD_SET(temp_accept->fd, &temp_f);
      nfds = (nfds > temp_accept->fd) ? nfds : temp_accept->fd;
      
    }

    for(accept_prev = NULL, temp_accept = fd_accept_list_head;
	temp_accept;
	accept_prev = temp_accept, temp_accept = temp_accept->next){

      if(FD_ISSET(temp_accept->fd, &f)){
	/*
	 * This means the appl has sent its first message.
	 */
#ifdef DEBUG
	printf("Appl's first message. Should be init. At time %u %u\n",
	       current_time.tv_sec, current_time.tv_usec);
#endif

	FD_CLR(temp_accept->fd, &f);
	n = recv(temp_accept->fd, recv_msg, MAX_MSG_SIZE, 0);

	if(n < 0)
	  err_dump("Error in recv! Exiting...");

	if(n == 0){
	  /*
	   * This means connection close. Take necessary steps.
	   * Here we just remove this fd from list and unset
	   * it from fd_set.
	   */
	  if(accept_prev != NULL)
	    accept_prev->next = temp_accept->next;
	  else
	    fd_accept_list_head = temp_accept->next;

	  FD_CLR(temp_accept->fd, &temp_f);
	  free(temp_accept);
	}
	else{
#ifdef DEBUG
	  print_packet(recv_msg, n);
#endif
	  handle_init(recv_msg, n, accept_prev, temp_accept);
	}
      }
    }

    for(init_prev = NULL, temp_init = fd_init_list_head;
	temp_init;
	init_prev = temp_init, temp_init = temp_init->next){
      
      REC_MULTICAST *temp_rec;

      for(temp_rec = temp_init->rec_list_head;
	  temp_rec;
	  temp_rec = temp_rec->next){
	if(FD_ISSET(temp_rec->fd, &f)){
	  /* A multicast packet for reconfigure has been received */
	  FD_CLR(temp_rec->fd, &f);
	  n = recv(temp_rec->fd, recv_msg, MAX_MSG_SIZE, 0);
	  
	  if(n < 0)
	    err_dump("Error in recv! Exiting...");
	  
	  if(n == 0){
	    /*
	     * This means connection close. Take necessary steps.
	     * This might just be a junk packet. Ignoring...
	     */
	  }
	  else{
	    /* Just forward this to appl */
	    if(*((uint8_t *) recv_msg) == DHCP_RECONFIGURE ||
	       *((uint8_t *) recv_msg) == DHCP_RECONFIGURE_INIT){
	      if(send(temp_init->tcp_fd, recv_msg, n, 0) != n)
		err_dump("Error in sending reconfigure message to appl. Exiting...");
	    }
	    else{
#ifdef DEBUG
	      printf("Message with invalid message type received on reconf multicast group\n");
#endif
	    }
	  }
	}
      }

      if(FD_ISSET(temp_init->tcp_fd, &f)){
	/* Appl has sent some request */
	FD_CLR(temp_init->tcp_fd, &f);
	n = recv(temp_init->tcp_fd, recv_msg, MAX_MSG_SIZE, 0);

	if(n < 0)
	  err_dump("Error in recv! Exiting...");

	if(n == 0){
	  /*
	   * This means connection close. Take necessary steps.
	   * Here we just remove this fd from list and unset
	   * it from fd_set.
	   */
	  printf("Connection close. Deleting %d\n", temp_init->pid);
	  if(init_prev != NULL)
	    init_prev->next = temp_init->next;
	  else
	    fd_init_list_head = temp_init->next;

	  FD_CLR(temp_init->tcp_fd, &temp_f);
	  free(temp_init);
	}
	else{
	  handle_request_from_appl(recv_msg, n, temp_init);
	}
      }
      if(FD_ISSET(temp_init->if_fd->fd, &f)){
	/* 
	 * Server has sent a reply. Note: If an appl. requests a
	 * server on some i/f and dies, it will be removed from
	 * init list. Now, if server sends a message to this addr
	 * it will not be consumed. Therefore, in next for loop
	 * we consume such messages.
	 */
	FD_CLR(temp_init->if_fd->fd, &f);
	server_socklen = sizeof(server_sockaddr);
	n = recvfrom(temp_init->if_fd->fd, 
		     recv_msg,
		     MAX_MSG_SIZE, 
		     0, 
		     (struct sockaddr *) &server_sockaddr, 
		     &server_socklen);

	print_packet(recv_msg, n);
	if(n < 0)
	  err_dump("Error in recv! Exiting...");

	if(n == 0){
	  /*
	   * Imp: This means connection close. Take necessary steps.
	   * Here we must remove this server from solicit list of
	   * this interface.
	   */
#ifdef DEBUG
	  printf("Server %s has closed connection\n", inet_ntop(AF_INET6, &server_sockaddr.sin6_addr, name, 127));
#endif
	}
	else{
	  handle_reply_from_server(recv_msg, n, &server_sockaddr);
	}
      }
    }
    /* Consume orphaned messages from servers or any malicious message */
    for(temp_interface = in_list_head; temp_interface; temp_interface = temp_interface->next){
      if(FD_ISSET(temp_interface->link_local_fd->fd, &f)){
	n = recvfrom(temp_interface->link_local_fd->fd, 
		     recv_msg,
		     MAX_MSG_SIZE, 
		     0, 
		     (struct sockaddr *) &server_sockaddr, 
		     &server_socklen);
	
	if(n < 0)
	  err_dump("Error in recv! Exiting...");
	
	if(n == 0){
	  /*
	   * Imp: This means connection close. Take necessary steps.
	   * Here we must remove this server from solicit list of
	   * this interface.
	   */
#ifdef DEBUG
	  printf("Server %s has closed connection\n", inet_ntop(AF_INET6, &server_sockaddr.sin6_addr, name, 127));
#endif
	}
	else{
	  printf("Orphaned message from server %s. No appl. alive/willing to recieve\n",
		 inet_ntop(AF_INET6, &server_sockaddr.sin6_addr, name, 127));
	}

      }

      for(temp_addr = temp_interface->ip_addr_fd_head; temp_addr; temp_addr = temp_addr->next){
	if(FD_ISSET(temp_addr->fd, &f)){
	  n = recvfrom(temp_addr->fd, 
		       recv_msg,
		       MAX_MSG_SIZE, 
		       0, 
		       (struct sockaddr *) &server_sockaddr, 
		       &server_socklen);
	  
	  if(n < 0)
	  err_dump("Error in recv! Exiting...");
	  
	  if(n == 0){
	    /*
	     * Imp: This means connection close. Take necessary steps.
	     * Here we must remove this server from solicit list of
	     * this interface.
	     */
#ifdef DEBUG
	    printf("Server %s has closed connection\n", inet_ntop(AF_INET6, &server_sockaddr.sin6_addr, name, 127));
#endif
	  }
	  else{
	    printf("Orphaned message from server %s. No appl. alive/willing to recieve\n",
		   inet_ntop(AF_INET6, &server_sockaddr.sin6_addr, name, 127));
	  }
	  
	}	
      }
      
    }
  
    temp_time = retransmit_timedout_messages();
    /* If you want the client daemon to delete expired leases do it here once again */

    if(temp_time.tv_sec == 0 && temp_time.tv_usec == 0){
#ifdef DEBUG
      printf("Sleeping again for %u microsecs\n", SLEEP_TIME * 100);
#endif
      SET_TIMEOUT(timeout, SLEEP_TIME * 100);
    }
    else{
      if((temp_int = DIFF_IN_MICRO(temp_time, current_time)) > 0){
#ifdef DEBUG
	printf("Sleeping for %u microsecs\n", temp_int);
#endif
	SET_TIMEOUT(timeout, temp_int);
	
      }
      else{
#ifdef DEBUG
	printf("Sleeping for 0 millisecs\n");
#endif
	SET_TIMEOUT(timeout, 0);
      }
    }
    f = temp_f;
  }
}

