void initialise_addr(IADDR *addr){
  /*
   * This function initializes all fields of addr structure to 0 or NULL.
   */
  bzero(addr, sizeof(*addr));
  inet_ntop(AF_INET6, &addr->ip_addr, "::", 3);
  inet_ntop(AF_INET6, &addr->server_addr, "::", 3);
  addr->preferred_lifetime = 0;
  addr->valid_lifetime     = 0;
  addr->next               = addr->prev = NULL;
}
/*********************************************************************************/

INTERFACE *find_interface(uint index, INTERFACE *head){
  /*
   * Searches the link list of interface structures for
   * an entry with interface index = index. head points
   * to the head of link list which is the first entry
   * in link list.
   */
  INTERFACE *tmp = head;

  while(tmp != NULL){
    if(tmp->index == index)
      break;
    tmp = tmp->next;
  }
  return tmp;
}
/*********************************************************************************/

INTERFACE *find_interface_with_linklocaladdr(uint index, INTERFACE *head){
  /*
   * Searches the link list of interface structures for
   * an entry with interface index = index. head points
   * to the head of link list which is the first entry
   * in link list.
   */
  INTERFACE *tmp = head;

  while(tmp != NULL){
    if(tmp->index == index && tmp->link_local_addr != NULL)
      break;
    tmp = tmp->next;
  }
  return tmp;
}
/*********************************************************************************/

void insert_in_list(IADDR *addr, IADDR **head){
             /* or (INTREFACE *addr, INTERFACE **head){ */
  /*
   * Inserts the structure pointed to by addr at the
   * start of the link list pointed to by *head. Note
   * that address of head is passed because head itself
   * may change due to insert.
   */
  if(addr == NULL){

#ifdef DEBUG
    printf("Nothing to insert into list\n");
#endif
    return;
  }

  if(*head != NULL){
    (*head)->prev = addr;
  }
 
  addr->next    = *head;
  *head         = addr;
  (*head)->prev = NULL;
}

/*********************************************************************************/
void delete_from_list(IADDR *addr, IADDR **head){
  /*
   * Deletes the structure pointed to by addr from link
   * list with head = *head and frees memory.
   */

  if(addr == NULL){

#ifdef DEBUG
    printf("Nothing to delete from list\n");
#endif
    return;
  }

  if(*head == NULL){

#ifdef DEBUG
    printf("Can't delete from empty list\n");
#endif
    return;
  }

  if(addr->prev != NULL)
    addr->prev->next = addr->next;

  if(addr->next != NULL)
    addr->next->prev = addr->prev;

  if(addr == *head)
    *head = (*head)->next;

  free(addr);
}
/*********************************************************************************/
