TIME retransmit_timedout_messages(){
  TIME     min;
  uint8_t *temp_msg_type, flag;

  if(gettimeofday(&current_time, NULL) == -1)
    err_dump("Error in gettimeofday. Exiting...");

  min.tv_sec  = 0;
  min.tv_usec = 0;
  flag        = 0;
  /* Scan init_list checking for timed out messages. Also get min next timeout to set 
     timeout for select */
  for(temp_init = fd_init_list_head; temp_init; temp_init = temp_init->next){
    if(temp_init->request_pending == 1){
      printf("Loop [%d] request pending for appl %d\n", loop, temp_init->pid);
      printf("next trans = %u %u current = %u %u\n",
	     (temp_init->next_transmission).tv_sec,
	     (temp_init->next_transmission).tv_usec,
	     (current_time).tv_sec,
	     (current_time).tv_usec);
      if(DIFF_IN_MICRO(temp_init->next_transmission, current_time) <= 0){
	/*
	 * retransmit message. Set next_transmission depending on message type.
	 * Also, set when_sent if no_of_attempts == 0 because this means message
	 * hasn't been transmitted till now.
	 */
	
	temp_msg_type = (uint8_t *)temp_init->msg;
	
	switch(*temp_msg_type){
	  
	case DHCP_SOLICIT:
	  if(temp_init->no_of_attempts > 0 && temp_init->adv_received){
	    {
	      TIME temp_time;
	      if(gettimeofday(&temp_time, NULL))
		errx("Error in gettimeofday. Exiting...\n");

	      printf("Sending advertise to appl at %u %u current was %u %u\n",
		     temp_time.tv_sec , temp_time.tv_usec,
		     current_time.tv_sec, current_time.tv_usec);
	    }
	    send_adv_to_appl(temp_init);
	    break;
	  }
	  
	  if(SOL_MAX_ATTEMPTS != -1){ /* not infinite no of retransmissions */
	    if(temp_init->no_of_attempts == SOL_MAX_ATTEMPTS){
	      /* no more retransmission of message. Return error to appl. */
	      /* Imp:Be Careful for solicit message bec you have to wait ADV_MSG_TIMEOUT
	       * to collect replies and return the reply to appl not error
	       */
	      send_error(temp_init, ALL_RETRANS_ATTEMPTS_FAILED);
	      temp_init->request_pending = 0;
	      break;
	    }
	  }
	  
	  if(temp_init->no_of_attempts == 0)
	    temp_init->when_sent = temp_init->next_transmission;
	  
	  
	  if(temp_init->no_of_attempts > 0){
	    temp_init->timeout *= 2;
	    if(temp_init->timeout > (ADV_MSG_MAX * 1000))
	      temp_init->timeout = (ADV_MSG_MAX * 1000);
	  }
	  //printf("****************Before adding %u %u ", (temp_init->next_transmission).tv_sec, (temp_init->next_transmission).tv_usec);
	  add_in_ms(&temp_init->next_transmission,
		    temp_init->timeout,
		    &temp_init->next_transmission);
	  
	  //printf("****************After adding %u %u ", (temp_init->next_transmission).tv_sec, (temp_init->next_transmission).tv_usec);
	  retransmit_msg(temp_init, SOL_MAX_ATTEMPTS);
	  break;
	  
	case DHCP_REQUEST:
	  
	  if(REQ_MSG_ATTEMPTS != -1){
	    if(temp_init->no_of_attempts == REQ_MSG_ATTEMPTS){
	      send_error(temp_init, ALL_RETRANS_ATTEMPTS_FAILED);
	      temp_init->request_pending = 0;
	      break;
	    }
	  }
	  else
	    printf("No of request message attempts shouldn't be infinite\n");
	  
	  temp_init->timeout *= 2;
	  add_in_ms(&temp_init->next_transmission, temp_init->timeout, &temp_init->next_transmission);
	  retransmit_msg(temp_init, REQ_MSG_ATTEMPTS);
	  break;
	  
	case DHCP_RELEASE:
	  
	  if(REL_MSG_ATTEMPTS != -1){
	    if(temp_init->no_of_attempts == REL_MSG_ATTEMPTS){
	      send_error(temp_init, ALL_RETRANS_ATTEMPTS_FAILED);
	      temp_init->request_pending = 0;
	      break;
	    }
	  }
	  else
	    printf("No of release message attempts shouldn't be infinite\n");
	  
	  temp_init->timeout *= 2;
	  add_in_ms(&temp_init->next_transmission, temp_init->timeout, &temp_init->next_transmission);
	  retransmit_msg(temp_init, REL_MSG_ATTEMPTS);
	  break;
	  
	case DHCP_RECONFIGURE_REPLY:
	  printf("Reconfigure Reply messages aren't retransmitted\n");
	  break;
	  
	default:
	  /* The message has invalid msg_type. This shouldn't happen*/
	  printf("Invalid message type found while retransmission\n");
	  break;
	}

	if(DIFF_IN_MICRO(temp_init->next_transmission, current_time) > 0){
	  if(flag == 0){
	    flag = 1;
	    min = temp_init->next_transmission;
	  }
	  else if(DIFF_IN_MICRO(temp_init->next_transmission, min) < 0){
	    min = temp_init->next_transmission;
	  }
	  
	}
      }
      else{
	/* DIFF > 0 */
	if(flag == 0){
	  flag = 1;
	  min = temp_init->next_transmission;
	}
	else if(DIFF_IN_MICRO(temp_init->next_transmission, min) < 0){
	  min = temp_init->next_transmission;
	}
      }
    }
  }
  return min;
}
