uint8_t init(){

  /*
   * This function is a part of library which applications 
   * will use.
   * Sending initialization message to client daemon on
   * behalf of the application. This would send
   * (1) msg_type   = 0
   * (2) pid        = process id of the application
   * (3) request_no = 1 beacuse this is start of connection
   */
  
  bzero((char *) &clientd_sockaddr, sizeof(clientd_sockaddr));

#ifdef SIN6_LEN
  clientd_sockaddr.sin6_len      = sizeof(clientd_sockaddr);      /* 4.3 BSD doesn't have this field      */
#endif

  clientd_sockaddr.sin6_family   = AF_INET6;                      /* Family as required for Ipv6          */
  clientd_sockaddr.sin6_port     = htons(CLIENTD_PORT);           /* Server listens on SERVER_PORT        */
  clientd_sockaddr.sin6_flowinfo = 0;                             /* Flow & Traffic Class are unimportant */
  clientd_sockaddr.sin6_scope_id = 0;                             /* Not really required.                 */
  clientd_sockaddr.sin6_addr     = in6addr_loopback;

  /* Open a TCP socket for the talking to client daemon   */
  sockfd                         = socket(AF_INET6, SOCK_STREAM, 0);

  if(sockfd < 0){
    printf("Error in opening datagram socket! Returning...\n");
    return 0;
  }

  /* Connect to the client daemon running on CLIENTD_PORT */

  if(connect(sockfd, (struct sockaddr *) &clientd_sockaddr, sizeof(clientd_sockaddr)) < 0){
    printf("Error in connecting to the client daemon. Returning...\n");
    return 0;
  }

  header.msg_type   = 0;
  header.pid        = pid = getpid();
  header.request_no = request_no++;

#ifdef DEBUG
  printf("Sending to client daemon: \nmsg_type   = %u \npid        = %u \nrequest_no = %u\n",
	 header.msg_type,
	 header.pid,
	 header.request_no);
#endif

  temp              = (uint32_t *) &header;
  *temp             = htonl(*temp);
  header.request_no = htonl(header.request_no);

  n                 = sizeof(header);
  memcpy(send_msg, &header, n);

  if(send(sockfd, send_msg, n, 0) != n){  //, (struct sockaddr *) &clientd_sockaddr, sizeof(clientd_sockaddr)
    printf("Error in sending initialization message to client daemon. Returning...\n");
    return 0;
  }

  n = recv(sockfd, recv_msg, MAX_MSG_SIZE, 0);
  if(n < 0){
    printf("Error in recv. Exiting...");
    return 0;
  }

  memcpy(&header, recv_msg, sizeof(header));
  temp               = (uint32_t *) &header;
  *temp              = ntohl(*temp);
  header.request_no  = ntohl(header.request_no);

#ifdef DEBUG
  printf("Reply from client daemon\n\n");
  printf("msg_type   = %u\n", header.msg_type);
  printf("pid        = %u\n", header.pid);
  printf("request_no = %u\n", header.request_no);
#endif
    
  return 1;
}
/************************************************************************/

uint8_t init_index(uint32_t if_index){

  /*
   * This function is a part of library which applications 
   * will use.
   * Sending initialization message to client daemon on
   * behalf of the application. This would send
   * (1) msg_type   = 0
   * (2) pid        = process id of the application
   * (3) request_no = 1 beacuse this is start of connection
   * (4) index      = if_index 
   *
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |         msg_type = 0          |        pid (16 bits)          |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                     request_no (32 bits)                      |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   * |                       index (32 bits)                         |
   * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   *
   */

  if(if_index == 0)
    return 0;

  index                          = if_index;

  bzero((char *) &clientd_sockaddr, sizeof(clientd_sockaddr));

#ifdef SIN6_LEN
  clientd_sockaddr.sin6_len      = sizeof(clientd_sockaddr);      /* 4.3 BSD doesn't have this field      */
#endif

  clientd_sockaddr.sin6_family   = AF_INET6;                      /* Family as required for Ipv6          */
  clientd_sockaddr.sin6_port     = htons(CLIENTD_PORT);           /* Server listens on SERVER_PORT        */
  clientd_sockaddr.sin6_flowinfo = 0;                             /* Flow & Traffic Class are unimportant */
  clientd_sockaddr.sin6_scope_id = 0;                             /* Not really required.                 */

  n = inet_pton(AF_INET6, CLIENTD_ADDRESS,                        /* Converting address to binary format. */
		& clientd_sockaddr.sin6_addr);                    /* Later you can change to loopback     */
  // clientd_sockaddr.sin6_addr = in6addr_loopback
  if(n != 1){
    if(n < 0)
      err_dump("Unknown AF(first) argument. Exiting...");
    else
      err_dump("Invalid address string. Exiting...");
  }


  /* Open a TCP socket for the talking to client daemon   */
  sockfd         = socket(AF_INET6, SOCK_STREAM, 0);

  if(sockfd < 0){
    printf("Error in opening datagram socket! Returning...\n");
    return 0;
  }

  /* Connect to the client daemon running on CLIENTD_PORT */

  if(connect(sockfd, (struct sockaddr *) &clientd_sockaddr, sizeof(clientd_sockaddr)) < 0){
    printf("Error in connecting to the client daemon. Returning...\n");
    return 0;
  }

  header.msg_type   = 0;
  header.pid        = pid = getpid();
  header.request_no = request_no++;
  header.index      = index;

#ifdef DEBUG
  printf("Sending to client daemon:");
  printf("\nmsg_type   = %u \npid        = %u \nrequest_no = %u \nindex      = %u \n",
	 header.msg_type,
	 header.pid,
	 header.request_no,
	 header.index);
#endif

  temp              = (uint32_t *) &header;
  *temp             = htonl(*temp);
  header.request_no = htonl(header.request_no);
  header.index      = htonl(header.index);

  n                 = sizeof(header);
  memcpy(send_msg, &header, n);

  if(send(sockfd, send_msg, n, 0) != n){
    printf("Error in sending initialization message to client daemon. Returning...\n");
    return 0;
  }

  n = recv(sockfd, recv_msg, MAX_MSG_SIZE, 0);
  if(n < 0){
    printf("Error in recv. Exiting...");
    return 0;
  }

  memcpy(&header, recv_msg, sizeof(header));
  temp               = (uint32_t *) &header;
  *temp              = ntohl(*temp);
  header.request_no  = ntohl(header.request_no);
  header.index       = ntohl(header.index);

#ifdef DEBUG
  printf("Reply from client daemon\n\n");
  printf("msg_type   = %u\n", header.msg_type);
  printf("pid        = %u\n", header.pid);
  printf("request_no = %u\n", header.request_no);
  printf("index      = %u\n", header.index);
#endif
 
  if(header.index == 0)
    /*
     * If header.index = 0 it means client daemon has sent an error.
     */
    return 0;
  
  return 1;
}
