SUBNET *find_subnet_with_prefix(struct in6_addr *prefix, int prefix_len){
  SUBNET               *sub;
  IN6_ADDR_WITH_PREFIX *pre;

  for(sub = subnet_list_head; sub; sub = sub->next){
    for(pre = sub->addr_list_head; pre; pre = pre->next){
      if(pre->prefix_len == prefix_len &&
	 IN6_ARE_PREFIXES_EQUAL(&pre->prefix, prefix, prefix_len))
	return sub;
    }
  }
  return NULL;
}

SUBNET *find_subnet_with_prefix_by_substring_match(struct in6_addr *prefix,
						   uint8_t *prefix_len) {
  /* prefix_len is set to length of prefix which has matched */
  SUBNET               *sub;
  IN6_ADDR_WITH_PREFIX *pre;

  if(!prefix_len)
    return NULL;

  for(sub = subnet_list_head; sub; sub = sub->next){
    for(pre = sub->addr_list_head; pre; pre = pre->next){
      if(IN6_ARE_PREFIXES_EQUAL(&pre->prefix, prefix, pre->prefix_len)) {
	*prefix_len = pre->prefix_len;
	return sub;
      }
    }
  }
  return NULL;
}

HOST *find_on_link_host(INTERFACE *interface, struct in6_addr *link_local_addr){
  HOST *host;
  
  for(host = interface->host_list_head; host; host = host->next)
    if(IN6_ARE_ADDR_EQUAL(&host->link_local_addr, link_local_addr))
      return host;

  return NULL;
}

HOST *find_off_link_host(SUBNET *subnet, struct in6_addr *link_local_addr){
  HOST *host;
  
  for(host = subnet->host_list_head; host; host = host->next){
    if(IN6_ARE_ADDR_EQUAL(&host->link_local_addr, link_local_addr))
      return host;
  }
  return NULL;
}

LEASE *find_lease(LEASE *head, struct in6_addr *link_local_addr,
		   struct in6_addr *relay_addr,	INTERFACE *interface,
		   struct in6_addr *ip_addr){
  /*
   * Return a lease with ip address = "ip_addr" leased to link-local-addr
   * "link_local_addr" with relay addr = "relay_addr" or interface name = 
   * interface->name.
   */
  LEASE *temp_lease;

  for(temp_lease = head; temp_lease; temp_lease = temp_lease->next){

    if(IN6_ARE_ADDR_EQUAL(link_local_addr, &temp_lease->leased_to) &&
       IN6_ARE_ADDR_EQUAL(ip_addr, &temp_lease->ip_addr)){

      if(relay_addr && !IN6_IS_ADDR_UNSPECIFIED(relay_addr)){
	if(IN6_ARE_ADDR_EQUAL(& (temp_lease->un).st.subnet_prefix, relay_addr))
	  return temp_lease;
      }
      else{
	if(strcmp((temp_lease->un).interface_name, interface->name) == 0)
	  return temp_lease;
      }
    }
  }
  return NULL;
}

LEASE **find_lease_addr(LEASE **head, struct in6_addr *link_local_addr,
			struct in6_addr *relay_addr, INTERFACE *interface,
			struct in6_addr *ip_addr){
  /*
   * Return a lease address with ip address = "ip_addr" leased to link-local-addr
   * "link_local_addr" with relay addr = "relay_addr" or interface name = 
   * interface->name.
   */
  LEASE **temp_lease;

  for(temp_lease = head; *temp_lease; temp_lease = &(*temp_lease)->next){

    if(IN6_ARE_ADDR_EQUAL(link_local_addr, &(*temp_lease)->leased_to) &&
       IN6_ARE_ADDR_EQUAL(ip_addr, &(*temp_lease)->ip_addr)){

      if(relay_addr && !IN6_IS_ADDR_UNSPECIFIED(relay_addr)){
	if(IN6_ARE_ADDR_EQUAL(& ((*temp_lease)->un).st.subnet_prefix, relay_addr))
	  return temp_lease;
      }
      else{
	if(strcmp(((*temp_lease)->un).interface_name, interface->name) == 0)
	  return temp_lease;
      }
    }
  }
  return NULL;
}
