# include "../common/type_len_structure.h"
# include "../common/extensions.h"
# include "generic_error.h"
# include "server_specific_error.h"
# include "find_subnet_interface_host.h"
# include "send_msg_to_client.h"
# include "get_new_addr.h"

#ifndef JUNK
#define JUNK
#endif

typedef struct lease_in_ext {
  LEASE *lease;
  struct lease_in_ext *next;
}LEASE_IN_EXT;

typedef LEASE_IN_EXT RETRANS_LEASE;

void build_lease_if_retransmission(RETRANS_LEASE **head, struct in6_addr *link_local_addr,
				   //struct in6_addr *relay_addr,
				   uint16_t transaction_id) {
  LEASE          *t_lease;

  /* Imp: Just uncomment all the single comments in this function if you
   * believe that retransmission check involves matching relay address
   * as well.
   */
  for(t_lease = lease_list_head; t_lease; t_lease = t_lease->next) {
    if(t_lease->xact_id == transaction_id &&
       //IN6_ARE_ADDR_EQUAL(&t_lease->relay_addr, relay_addr) &&
       IN6_ARE_ADDR_EQUAL(&t_lease->link_local_addr, link_local_addr)) {

      *head = (RETRANS_LEASE *) malloc(sizeof(RETRANS_LEASE));
      if(!(*head))
	errx(1, "Insufficient memory. Exiting...\n");
      bzero((char *) *head, sizeof(**head));

      (*head)->lease = t_lease;
      head           = &(*head)->next
    }
  }
}

int has_site_local_addr(INTERFACE *head) {
  IN6_ADDR_LIST *temp_iaddr;

  for(; head != NULL; head = head->next) {
    for(temp_iaddr = head->addr_list_head; temp_iaddr; temp_iaddr = temp_iaddr->next) {
      if(IN6_IS_ADDR_SITELOCAL(&temp_iaddr->ip_addr))
	return 1;
    }
  }
  return 0;
}

int has_global_addr(INTERFACE *head) {
  IN6_ADDR_LIST *temp_iaddr;
  
  for(; head != NULL; head = head->next) {
    for(temp_iaddr = head->addr_list_head; temp_iaddr; temp_iaddr = temp_iaddr->next) {
      if(IN6_IS_ADDR_GLOBAL(&temp_iaddr->ip_addr))
	return 1;
    }
  }
  return 0;
}

int has_addr(struct in6_addr *addr, INTERFACE *head) {
  IN6_ADDR_LIST *temp_iaddr;
  
  for(; head != NULL; head = head->next) {
    for(temp_iaddr = head->addr_list_head; temp_iaddr; temp_iaddr = temp_iaddr->next) {
      if(IN6_ARE_ADDR_EQUAL(&temp_iaddr->ip_addr, addr))
	return 1;
    }
  }
  return 0;
}

int check_scope(struct in6_addr *relay_addr) {

  /* Now check scope of relay address */
  /* You can use what Sir told. Ping relay to validate scope as well as to 
   * find addr to use for solicit */
  
  if(IN6_IS_ADDR_MULTICAST(relay_addr)) {

#ifdef DEBUG
    printf("Relay address in client's message is multicast address."
	   " Discarding message\n");
#endif

    return 0;
  }
  if(IN6_IS_ADDR_LINKLOCAL(relay_addr) ||
     IN6_IS_ADDR_LOOPBACK(relay_addr)) {

#ifdef DEBUG
    printf("Invalid relay address in client's message. Discarding message...\n");
#endif
    
    return 0;
  }
  if(IN6_IS_ADDR_SITELOCAL(relay_addr)) {
    /*
     * Server must have a sitelocal or global address
     * But you only need to check. kernel wil automatically use an
     * appropriate address for replying because reply_fd is bound to
     * in6addr_any. The same holds for else below
     */
    if(!(has_site_local_addr(in_list_head) || has_global_addr(in_list_head))) {

#ifdef DEBUG
      printf("The server doesn't have a site local or global address to reply to"
	     " a sitelocal relay\n");
#endif

      return 0;
    }
  }
  else {
    /* Server must have a global address */
    if(!has_global_addr(in_list_head)) {

#ifdef DEBUG
      printf("The server doesn't have a global address to reply to"
	     " a global relay\n");
#endif

      return 0;
    }
  }

  return 1;
}

int is_present(LEASE *temp_lease, LEASE_IN_EXT *head) {
  for(; head; head = head->next)
    if(head->lease == temp_lease)
      return 1;

  return 0;
}

uint32_t write_ip_address_extension(LEASE *lease, uint32_t length_written,
				    uint8_t status, uint8_t A, uint8_t P) {
  /* Returns the length written */
  /* A and P in argument tell whether client has asked for these 
   * updates. If the client hasn't asked for update we must not 
   * set A and P bits using values from lease->A and lease->P
   * because these might be results of previous updates. ex. is
   * if the current request is lease renewal without dns update.
   */
  TYPE_LEN tl;
  uint32_t old_length = length_written;
  uint32_t temp_ext;

  tl.type = htons(IP_ADDRESS_EXTENSION);
  /* Note: I am not writing tl because I don't know the length of the
   * extension. tl will be written at the end
   */
  length_written += sizeof(tl);

  temp_ext = 0;
  SET_BITS_HIGH(temp_ext, status, 8, 32);
  SET_BITS_MIDDLE(temp_ext, 1, 24, 1);
  SET_BITS_MIDDLE(temp_ext, 0, 23, 1);
  SET_BITS_MIDDLE(temp_ext, 1, 22, 1);
  SET_BITS_MIDDLE(temp_ext, 0, 21, 1);

  if(A && lease->A)
    SET_BITS_MIDDLE(temp_ext, 1, 20, 1);
  if(P && lease->P)
    SET_BITS_MIDDLE(temp_ext, 1, 19, 1);

  SET_BITS_MIDDLE(temp_ext, lease->scope, 10, 3);
  SET_BITS_MIDDLE(temp_ext, lease->ip_addr_prefix_len, 7, 7);
  
  temp_ext = htonl(temp_ext);
  lease->preferred_lifetime = htonl(lease->preferred_lifetime);
  lease->valid_lifetime = htonl(lease->valid_lifetime);
  WRITE_DATA(send_msg, length_written, &temp_ext, sizeof(temp_ext));
  WRITE_DATA(send_msg, length_written, &lease->ip_addr, sizeof(lease->ip_addr));
  WRITE_DATA(send_msg, length_written, &lease->preferred_lifetime,
	     sizeof(lease->preferred_lifetime));
  WRITE_DATA(send_msg, length_written, &lease->valid_lifetime,
	     sizeof(lease->valid_lifetime));
  lease->preferred_lifetime = ntohl(lease->preferred_lifetime);
  lease->valid_lifetime = ntohl(lease->valid_lifetime);

  tl.length = htons(length_written - old_length - sizeof(tl));
  memcpy(send_msg + old_length, &tl, sizeof(tl));
  return length_written;
}

uint32_t write_ip_address_extension_non_dhcp(uint32_t length_written,
					 struct in6_addr *ip_addr,
					 uint8_t status, uint8_t A, uint8_t P,
					 uint8_t scope, uint8_t prefix_len) {
  /* Note that A and P arguments of this function are results of dns update
   * not whether client has asked for update as in above function
   */

  TYPE_LEN tl;
  uint32_t old_length = length_written;
  uint32_t temp_ext;
  
  tl.type = htons(IP_ADDRESS_EXTENSION);
  /* Note: I am not writing tl because I don't know the length of the
   * extension. tl will be written at the end
   */
  length_written += sizeof(tl);
  temp_ext = 0;

  SET_BITS_HIGH(temp_ext, status, 8, 32);
  SET_BITS_MIDDLE(temp_ext, 1, 24, 1);/* Address is present */
  SET_BITS_MIDDLE(temp_ext, 1, 23, 1);/* non-dhcp address   */
  SET_BITS_MIDDLE(temp_ext, 0, 22, 1);/* no lifetimes       */
  SET_BITS_MIDDLE(temp_ext, 0, 21, 1);/* Q = 0 for reply    */
  SET_BITS_MIDDLE(temp_ext, A, 20, 1);/* A argument of fn   */
  SET_BITS_MIDDLE(temp_ext, P, 19, 1);/* P argument of fn   */
  SET_BITS_MIDDLE(temp_ext, scope, 10, 3);    /* scope arg of fn as sent by client */
  SET_BITS_MIDDLE(temp_ext, prefix_len, 7, 7);/* prefix-len as sent by client      */
  
  temp_ext = htonl(temp_ext);
  WRITE_DATA(send_msg, length_written, &temp_ext, sizeof(temp_ext));
  WRITE_DATA(send_msg, length_written, ip_addr, sizeof(*ip_addr));

  tl.length = htons(length_written - old_length - sizeof(tl));
  memcpy(send_msg + old_length, &tl, sizeof(tl));
}

void handle_packet_from_client(uint index, char *msg, int len,
			       struct sockaddr_in6 *client)//, int type_of_packet)
{
  uint8_t   *msg_type;
  uint32_t   length_read    = 0;
  uint32_t   length_written = 0;
  TYPE_LEN   tl;
  char       name[64];
  INTERFACE *interface;
  
  interface = find_interface(index, in_list_head);
  if(!interface)
    errx("Impossible. Incoming interface not found\n");

  msg_type = (uint8_t *)msg;
  
  switch(*msg_type) {
    
  case DHCP_SOLICIT:
    {
      SOLICIT       sol;
      ADVERTISE     adv;
      uint8_t       c, p, prefix_len, relay = 0;
      uint16_t      solicit_id;
      SUBNET       *subnet;
      HOST         *host;
      PREFIX_LIST  *temp_pre;
      LEASE       **temp_lease, *free_lease;
      
      if(len != sizeof(sol)) {
	printf("Incorrect sized solicit message. Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      READ_DATA(&sol, msg, length_read, sizeof(sol));
      sol.prefix_len_solicit_id = ntohs(sol.prefix_len_solicit_id);
      
      c          = GET_BITS_HIGH(sol.c_p_reserved, 1, 8);
      p          = GET_BITS_MIDDLE(sol.c_p_reserved, 7, 1);
      prefix_len = GET_BITS_HIGH(sol.prefix_len_solicit_id, 7, 16);
      solicit_id = GET_BITS_LOW(sol.prefix_len_solicit_id, 9);
#ifdef PACKET
      printf("Solicit.........................\n");
      printf("c = %u p = %u prefix-len = %u solicit_id = %u", c, p, prefix_len, solicit_id);
#endif
      if(!IN6_IS_ADDR_LINKLOCAL(&sol.client_link_local_addr)) {
	printf("Client's address specified in solicit message isn't link local."
	       " Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      

      if(!IN6_IS_ADDR_UNSPECIFIED(&sol.relay_addr)) {
	relay = 1;
	
	if(prefix_len == 0) {
	  printf("Non zero relay address but prefix len is zero."
		 " Discarding message\n");
	  break;/* break out of switch. You can use return as well */
	}
	if(!check_scope(&sol.relay_addr))
	  break;
      }
      else {
	/* Imp: get the link local addr of incoming interface. Need not do
	   because you specify outgoing if in dest scope id */
      }

      if(relay) {
	subnet = find_subnet_with_prefix(&sol.relay_addr, prefix_len);
	if(!subnet) {
	  printf("Solicit from unknown subnet. Discarding message\n");
	}

	host = find_off_link_host(subnet, &sol.client_link_local_addr);
	if(host) {
	  adv.preference  = host->preference;
	  adv.server_addr = (host->server_addr).prefix;
	  temp_pre        = host->adv_prefix_list_head;
	}
	else {
	  adv.preference  = subnet->preference;
	  adv.server_addr = (subnet->server_addr).prefix;
	  temp_pre        = subnet->adv_prefix_list_head;
	}

      }
      else {
	host = find_on_link_host(interface, &sol.client_link_local_addr);
	if(host) {
	  adv.preference  = host->preference;
	  adv.server_addr = (host->server_addr).prefix;
	  temp_pre        = host->adv_prefix_list_head;
	}
	else {
	  adv.preference  = interface->preference;
	  adv.server_addr = (interface->server_addr).prefix;
	  temp_pre        = interface->adv_prefix_list_head;
	}
      }
      if(c) {
	/* Imp: Clear all entries */
	temp_lease = &lease_list_head;
	while(*temp_lease) {
	  if(IN6_ARE_ADDR_EQUAL(&sol.client_link_local_addr,
				&(*temp_lease)->leased_to)) {
	    if(relay) {
	      if(prefix_len == ((*temp_lease)->un).st.subnet_prefix_len &&
		 IN6_ARE_ADDR_EQUAL(&sol.relay_addr, &((*temp_lease)->un).st.subnet_prefix)) {
		/* Note: I have matched whole address not just prefix
		 * because binding should be <client-link-local-addr,
		 * relay-addr not prefix>
		 */
		free_lease  = *temp_lease;
		*temp_lease = (*temp_lease)->next;

		/* Imp: update lease file */
		free(free_lease);
		continue;
	      }
	    }
	    else {
	      if(strcmp(((*temp_lease)->un).interface_name, interface->name) == 0) {
		free_lease  = *temp_lease;
		*temp_lease = (*temp_lease)->next;

		/* Imp: update lease file */
		free(free_lease);
		continue;
	      }
	    }
	  }
	  temp_lease = &(*temp_lease)->next;
	}
      }

      adv.msg_type               = DHCP_ADVERTISE;
      adv.reserved_solicit_id[0] = GET_BITS_MIDDLE(sol.prefix_len_solicit_id, 9, 1);
      adv.reserved_solicit_id[1] = GET_BITS_LOW(sol.prefix_len_solicit_id, 8);
      adv.client_link_local_addr = sol.client_link_local_addr;
      adv.relay_addr             = sol.relay_addr;
      WRITE_DATA(send_msg, length_written, &adv, sizeof(adv));

      if(p) {
	tl.type = htons(SUBNET_PREFIX_EXTENSION);
	/* Attach prefix extensions and update length_written */
	for(;temp_pre; temp_pre = temp_pre->next) {
	  /* Note temp_pre has been initialised above */
	  tl.length = htons(1 + (temp_pre->prefix_len + 7)/8);
	  WRITE_DATA(send_msg, length_written, &tl, sizeof(tl));
	  WRITE_DATA(send_msg, length_written, &temp_pre->prefix_len,
		     sizeof(temp_pre->prefix_len));
	  WRITE_DATA(send_msg, length_written, &temp_pre->prefix,
		     ntohs(tl.length) - 1);
	}
      }

      client->sin6_port          = htons(CLIENT_PORT);
      if(relay) {
	client->sin6_addr        = sol.relay_addr;
	client->sin6_scope_id    = 0;
      }
      else {
	client->sin6_addr        = sol.client_link_local_addr;
	client->sin6_scope_id    = index;
      }
      print_packet(send_msg, length_written);
      getchar();
      send_msg_to_client(send_msg, length_written, client, index);
    }
    break;

  case DHCP_REQUEST:
    {
      REQUEST req;
      REPLY   rep;
      SUBNET *subnet;
      uint8_t c, r, relay = 0, error = 0, status = 0, prefix_len, t_error;
      uint32_t rep_header_size;
      /* status is the status value to send in reply message */

      typedef struct dns_update {

	union {
	  LEASE            *lease;
	  struct {
	    char            dns_name[MAX_DNS_NAME];
	    struct in6_addr ip_addr;
	  }st;
	}un;

	uint8_t            A        :1;
	uint8_t            P        :1;
	uint8_t            operation:1;
	uint8_t            dhcp     :1;
	uint8_t            A_success:1;
	uint8_t            P_success:1;
	struct dns_update *next;
	
      }DNS_UPDATE;
      
      LEASE_IN_EXT    *lie_head = NULL, **temp_lie = &lie_head, *t_lie;
      LEASE           *t_lease, *t_lease_list_head = NULL;
      LEASE          **temp_lease = &t_lease_list_head;
      DNS_UPDATE      *t_dns, *dns_list_head = NULL, **temp_dns = &dns_list_head;
      RETRANS_LEASE   *rl_head = NULL, **temp_rl = &rl_head, *t_rl;
      
#ifdef JUNK
      printf("0\n");
#endif
      if(len < sizeof(req)) {
	printf("Incorrect sized request message. Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      READ_DATA(&req, msg, length_read, sizeof(req));
      req.transaction_id = ntohs(req.transaction_id);

      c = GET_BITS_HIGH(req.c_r_reserved, 1, 8);
      r = GET_BITS_MIDDLE(req.c_r_reserved, 7, 1);

      if(!IN6_IS_ADDR_LINKLOCAL(&req.client_link_local_addr)) {
	printf("Client's address specified in request message isn't link local."
	       " Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      
      if(!has_addr(&req.server_addr, in_list_head)) {
	printf("Couldn't find server's address specified in request message."
	       " Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }

      if(!IN6_IS_ADDR_UNSPECIFIED(&req.relay_addr)) {
	relay = 1;
	if(!check_scope(&req.relay_addr))
	  break;
      }
      if(relay) {
	subnet = find_subnet_with_prefix_by_substring_match(&req.relay_addr,
							    &prefix_len);
	/* prefix_len is set by the function call if a prefix match is found */
	if(!subnet) {
	  printf("Request from unknown subnet. Discarding message...\n");
	  break;
	}
      }
      else
	subnet = NULL;
      /* Imp: What about authentication ext? See page 38/67 */

      rep.msg_type = DHCP_REPLY;
      rep.r_status = 0;
      SET_BITS_HIGH(rep.r_status, relay, 1, 8);/* r = relay */
      rep.transaction_id         = htons(req.transaction_id);
      rep.client_link_local_addr = req.client_link_local_addr;
      /*
       * Imp: I am not writing rep into send_msg but updating length_written 
       * so that I can attcah extensions as required and later write rep. I
       * am not writing now because status field can't be filled in without
       * processing extensions
       */
      length_written += sizeof(rep);

      if(relay)
	/* Add relay address to reply message */
	WRITE_DATA(send_msg, length_written, &req.relay_addr, sizeof(req.relay_addr));

      rep_header_size = length_written;

      client->sin6_port          = htons(CLIENT_PORT);
      if(relay) {
	client->sin6_addr        = req.relay_addr;
	client->sin6_scope_id    = 0;
      }
      else {
	client->sin6_addr        = req.client_link_local_addr;
	client->sin6_scope_id    = index;
      }
      build_list_if_retransmission(&rl_head, &req.client_link_local_addr,//&req.relay_addr,
				   req.transaction_id);
      if(!rl_head) {
	printf("Retransmitted message...\n");
      }

      while(length_read < len) {
	READ_DATA(&tl, msg, length_read, sizeof(tl));
	tl.type      = ntohs(tl.type);
	tl.length    = ntohs(tl.length);
	error        = 0;

	switch(tl.type) {
	  
	case IP_ADDRESS_EXTENSION:
	  {
	    struct{
	      uint32_t status     : 8;
	      uint32_t C          : 1;
	      uint32_t I          : 1;
	      uint32_t L          : 1;
	      uint32_t Q          : 1;
	      uint32_t A          : 1;
	      uint32_t P          : 1;
	      uint32_t reserved   : 8;
	      uint32_t scope      : 3;
	      uint32_t prefix_len : 7;
	    } s;
	    
	    uint32_t         temp_ext;
	    struct in6_addr  ip_addr;
	    uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	    char             dns_name[MAX_DNS_NAME];
 
	    name_size    = 0;
	    dns_name[0]  = '\0';
	    old_length   = length_read;
	    READ_DATA(&temp_ext, msg, length_read, sizeof(temp_ext));
	    
	    temp_ext     = ntohl(temp_ext);
	    s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	    s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	    s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	    s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	    s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	    s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	    s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	    s.reserved   = GET_BITS_MIDDLE(temp_ext, 18, 8);
	    s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	    s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	    
	    if(s.C)
	      READ_DATA(&ip_addr, msg, length_read, sizeof(ip_addr));
	    
	    if(s.L) {
	      READ_DATA(& preferred_lifetime, msg, length_read, sizeof(preferred_lifetime));
	      READ_DATA(& valid_lifetime, msg, length_read, sizeof(valid_lifetime));
	      preferred_lifetime = ntohl(preferred_lifetime);
	      valid_lifetime     = ntohl(valid_lifetime);
	    }
	    
	    if((length_read - old_length) < tl.length) {
	      /* This means DNS name is present. Read it off */
	      name_size           = tl.length - (length_read - old_length);
	      READ_DATA(dns_name, msg, length_read, name_size);
	      dns_name[name_size] = '\0';
	      /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	      printf("DNS name = %s in IP ADDRESS EXTENSION\n", dns_name);
#endif
	      
	    }
	    if(s.reserved != 0 || 
	       s.status   != 0 || 
	       (!s.C && s.I) ||
	       (s.L && s.I)) {
	      /* (1) reserved MUST be zero.
	       * (2) In a DHCP Request message (for each IP address
	       *     extension), a client MUST initialize the ``status'' 
	       *     field value to zero.
	       *
	       * (3) The ``I'' bit MUST NOT be set if the ``C'' bit is not set
	       * (4) The ``L'' bit MUST NOT be set if the ``I'' bit is set
	       */

	      /* Send the incorrect extension with status field set to 2 */
	      error  = EXTENSION_SETTINGS_ILLEGAL;
	      if(status == 0)
		status = PoorlyFormed;
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;/* break out of switch to while loop */	      
	    }
	    if(r) {
	      /* Imp: Ask Sir */
	    }
	    if(c) {
	      /*
	       * The client requests to remove all releasable resources
	       * except those provided as extensions. We build a list of all
	       * the releasable resources associated with the client which are
	       * not to be freed.
	       */
	      if(!s.C || s.I || s.Q || s.A || s.P) {
		/* (1) !s.C => Client has asked to free everything except those in ext but
		 * extension doesn't contain address. This means client is asking
		 * for a new address. But then c bit in req shouldn't have been
		 * set.
		 * (2) s.I => asking for not releasing a non DHCP addr. But this doesn't
		 * make sense as non DHCP addr would never be released.
		 * (3) s.Q => This bit is used only while requesting lifetime of a non
		 * DHCP addr or updating DNS, not while asking for not releasing something.
		 * (4) s.A => used only for updating dns.
		 * (5) s.P => used only for updating dns.
		 */
		/* Send the incorrect extension with status field set to 2 */
#ifdef DEBUG
		printf("clear bit set and !s.C || s.I || s.Q || s.A || s.P\n");
#endif
		error = EXTENSION_SETTINGS_ILLEGAL;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;/* break out of switch to while loop */
	      }
	      t_lease = find_lease(lease_list_head, &req.client_link_local_addr,
				   &req.relay_addr, interface, &ip_addr);
	      
	      if(!t_lease) {
#ifdef DEBUG
		printf("Client has asked to release everything except a list containing %s. "
		       "But this address isn't present in server lease database\n",
		       inet_ntop(AF_INET6, &ip_addr, name, 64));
#endif

		/* Send error back: IP address requested by client is not available */
		error = REQUESTED_IP_ADDR_UNAVAILABLE;
		if(status == 0)
		  status = NoBinding;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;
	      }
	      *temp_lie          = (LEASE_IN_EXT *)malloc(sizeof(LEASE_IN_EXT));
	      (*temp_lie)->lease = t_lease;
	      temp_lie           = &(*temp_lie)->next;
	    }
	    else {
	      /* The client is asking for a new ip_addr, a specific addr, lease-renewal
	       * or dns update.
	       */
	      /* Imp: check if all these checks below can be moved above "if(c)" */
	      if(!s.C && s.prefix_len != 0) {
		/* if the ``C'' bit is not set, prefix-len MUST be zero */
#ifdef DEBUG
		printf("If the ``C'' bit is not set, prefix-len MUST be zero\n");
#endif
		error = EXTENSION_SETTINGS_ILLEGAL;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;
	      }

	      if((s.A || s.P) && (dns_name[0] == '\0')) {
		/* Client hasn't sent domain name to use in dns update */
#ifdef DEBUG
		printf("Client hasn't sent domain name to use in dns update\n");
#endif
		error = EXTENSION_SETTINGS_ILLEGAL;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;
	      }

	      if(!s.I && s.Q && !s.A  && !s.P) {
		/* If I = 0, Q bit is set only for dns updates but A = P = 0 => error */
#ifdef DEBUG
		printf("If I = 0, Q bit is set only for dns updates but A = P = 0 => error\n");
#endif
		error = EXTENSION_SETTINGS_ILLEGAL;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;
	      }
	      if(s.C && s.I && !(s.A || s.P)) {
		/* If I bit is set only dns update is possible. Therefore, either
		 * A or P MUST be 1
		 */
#ifdef DEBUG
		printf("If I bit is set only dns update is possible. "
		       "Therefore, either A or P bit MUST be 1\n");
#endif
		error = EXTENSION_SETTINGS_ILLEGAL;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;

	      }
	      
	      if(!s.C && !s.I) {
		/* Client has asked for lease of a new ip address */
		/* If no IP address field is present, then the client is
		 * requesting that the server allocate an IP address of the
		 * scope identified by the "scope" field value.
		 */

		/* Imp: Check for retransmission */
#ifdef DEBUG
		printf("Client has asked for lease of a new ip address\n");
#endif
		t_lease = (LEASE *) malloc(sizeof(LEASE));
		if(!t_lease)
		  errx(1, "Insufficient memory. Exiting...\n");
		bzero((char *) t_lease, sizeof(*t_lease));
		
		t_lease->leased_to = req.client_link_local_addr;

		if(relay) {
		  t_lease->onlink                    = 0;
		  (t_lease->un).st.subnet_prefix     = req.relay_addr;
		  (t_lease->un).st.subnet_prefix_len = prefix_len;
		}
		else {
		  t_lease->onlink                    = 1;
		  strcpy((t_lease->un).interface_name, interface->name);
		}

		t_lease->scope                = s.scope;
		t_lease->xact_id              = req.transaction_id;

		if(s.L) {
		  t_lease->preferred_lifetime = preferred_lifetime;
		  t_lease->valid_lifetime     = valid_lifetime;
		}
		else {
		  t_lease->preferred_lifetime = 0;
		  t_lease->valid_lifetime     = 0;
		}
		strcpy(t_lease->dns_name, dns_name);
		t_lease->abandoned = 0;
		/* Note the last argument of the following function call.
		 * It is a pointer to current temporary lease list head so that there
		 * are no duplicates if there are multiple ip address extensions in the same
		 * request message.
		 */
		error = get_new_addr_by_scope(subnet, interface, t_lease, s.A, s.P,
					      t_lease_list_head);
		
		if(error != REQUEST_GRANTED_NO_ERRORS) {
		  /* we can't continue further with this extension */
#ifdef DEBUG
		printf("Request couldn't be granted by get_new_addr_by_scope\n");
#endif
		  if(status == 0)
		    status = Unavail;
		  memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			 sizeof(tl) + tl.length);
		  memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		  length_written += (sizeof(tl) + tl.length);
		  break;
		}

		if(s.Q && s.A) {
		  /* error will be first error encountered */
		  t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, A_RECORD);
		  if(error == 0)
		    error = t_error;
		  if(t_error == 0)
		    t_lease->A = 1;
		  else if(status == 0)
		    status = UnspecFail;
		}

		if(s.Q && s.P) {
		  /* error will be first error encountered */
		  t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, P_RECORD);
		  if(error == 0)
		    error = t_error;
		  if(t_error == 0)
		    t_lease->P = 1;
		  else if(status == 0)
		    status = UnspecFail;
		}

		if(!s.Q && (s.A || s.P)) {
		  /* Build a list of dns updates to be made later */
		  *temp_dns              = (DNS_UPDATE *) malloc(sizeof(DNS_UPDATE));
		  if(!*temp_dns)
		    errx(1, "Insufficient memory. Exiting...\n");
		  bzero((char *)*temp_dns, sizeof(**temp_dns));

		  ((*temp_dns)->un).lease = t_lease;
		  (*temp_dns)->A          = s.A;
		  (*temp_dns)->P          = s.P;
		  (*temp_dns)->operation  = ADD;
		  (*temp_dns)->dhcp       = 1;
		  temp_dns                = &(*temp_dns)->next;
		}
		/* Write the extension with result in reply */
		length_written = write_ip_address_extension(t_lease, length_written,
							    error, s.A, s.P);
		*temp_lease = t_lease;
		temp_lease  = &(*temp_lease)->next;
	      }
	      else if(s.C && !s.I) {
		/* Client has asked for lease of a specific addr or lease renewal. */
		/* If the IP address field is present and the "I" bit is not set, then
		 * the client is requesting the assignment of:
		 *   (1) A specific IP address (interface ID present)
		 *   (2) Any IP address in the specified network (interface ID is zero)
		 */
		t_lease = find_lease(lease_list_head, &req.client_link_local_addr,
				     &req.relay_addr, interface, &ip_addr);
		if(t_lease) {
		  /*
		   * This is a lease renewal request. Currently I am granting renewal but
		   * you can change later depending upon your policy regarding this client
		   * by changing the function renewal_allowed(). Note that the only argument
		   * to this function is the lease because it contains all necessary information
		   * like client's link local address, relay/interface_index, when_leased etc.
		   * It doesn't make a difference whether lease has expired for the above if()
		   * as it would be taken care of by renewal_allowed().
		   */
		  if(!renewal_allowed(t_lease, has_expired(t_lease))) {
		    error = REQUESTED_IP_ADDR_UNAVAILABLE;
		    if(status == 0)
		      status = Unavail;
		    memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			   sizeof(tl) + tl.length);
		    memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		    length_written += (sizeof(tl) + tl.length);
		    /* Continue processing further extensions */
		    break;
		    
		  }

		  if(s.A && !t_lease->dns_update_A_allowed ||
		     s.P && !t_lease->dns_update_P_allowed) {
#ifdef DEBUG
		printf("lease renewal request: dns update doesn't permit request "
		       "to be granted\n");
#endif
		    error = REQUESTED_IP_ADDR_UNAVAILABLE;
		    if(status == 0)
		      status = Unavail;
		    memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			   sizeof(tl) + tl.length);
		    memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		    length_written += (sizeof(tl) + tl.length);
		    /* Continue processing further extensions */
		    break;
		  }
		  
		  if(s.L) {
		    /*
		     * Check whether the demanded lifetimes are within limits specified
		     * by their max values.
		     */
		    if(preferred_lifetime <= t_lease->max_preferred_lifetime &&
		       valid_lifetime <= t_lease->max_valid_lifetime) {
		      
		      t_lease->preferred_lifetime = preferred_lifetime;
		      t_lease->valid_lifetime     = valid_lifetime;
		      // if(gettimeofday(&t_lease->when_leased, NULL) == -1)
		      // errx(1, "Error in gettimeofday. Exiting...\n");
		    }
		    else {
#ifdef DEBUG
		printf("Requested lifetimes aren't available\n");
#endif
		      error = REQUESTED_IP_ADDR_UNAVAILABLE;
		      if(status == 0)
			status = Unavail;
		      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			     sizeof(tl) + tl.length);
		      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		      length_written += (sizeof(tl) + tl.length);
		      /* Continue processing further extensions */
		      break;
		    }
		  }
		  else {
		    t_lease->preferred_lifetime = t_lease->default_preferred_lifetime;
		    t_lease->valid_lifetime     = t_lease->default_valid_lifetime;
		  }

		  if(s.A || s.P)
		    strcpy(t_lease->dns_name, dns_name);
		  
		  if(s.Q && s.A) {
		    /* Note that I am not checking whether a dns update has already
		     * been made because it will be overwritten by current one. If you
		     * explicitly need to delete the old dns record, move the above
		     * strcpy here and before the strcpy place
		     * update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD);
		     * The same holds for P record below
		     */
		    t_lease->A = 0;
		    t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, A_RECORD);
		    if(error == 0)
		      error = t_error;
		    if(t_error == 0)
		      t_lease->A = 1;
		    else if(status == 0)
		      status = UnspecFail;
		  }
		  
		  if(s.Q && s.P) {
		    t_lease->P = 0;
		    t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, P_RECORD);
		    if(error == 0)
		      error = t_error;
		    if(t_error == 0)
		      t_lease->P = 1;
		    else if(status == 0)
		      status = UnspecFail;
		  }
		  
		  if(!s.Q && (s.A || s.P)) {
		    /* Build a list of dns updates to be made later */
		    *temp_dns              = (DNS_UPDATE *) malloc(sizeof(DNS_UPDATE));
		    if(!*temp_dns)
		      errx(1, "Insufficient memory. Exiting...\n");
		    bzero((char *)*temp_dns, sizeof(**temp_dns));
		    
		    ((*temp_dns)->un).lease = t_lease;
		    (*temp_dns)->A          = s.A;
		    (*temp_dns)->P          = s.P;
		    (*temp_dns)->operation  = ADD;
		    (*temp_dns)->dhcp       = 1;
		    temp_dns                = &(*temp_dns)->next;
		  }
		  t_lease->old       = 1;
		  t_lease->xact_id   = req.transaction_id;
		  /* Write the extension with result in reply */
		  length_written     = write_ip_address_extension(t_lease, length_written,
								  error, s.A, s.P);
		  *temp_lie          = (LEASE_IN_EXT *)malloc(sizeof(LEASE_IN_EXT));
		  (*temp_lie)->lease = t_lease;
		  temp_lie           = &(*temp_lie)->next;
		  
		} // if(t_lease)
		else {
		  /* Lease of a specific addr      */
		  /* Imp: Check for retransmission */
		  if(has_been_leased(&ip_addr, t_lease_list_head)) {
#ifdef DEBUG
		printf("lease of a specific addr: addr is in use\n");
#endif
		    error = IP_ADDRESS_IN_USE;
		    if(status == 0)
		      status = Unavail;
		    memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			   sizeof(tl) + tl.length);
		    memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		    length_written += (sizeof(tl) + tl.length);
		    /* Continue processing further extensions */
		    break;
		  }

		  t_lease = (LEASE *) malloc(sizeof(LEASE));
		  if(!t_lease)
		    errx(1, "Insufficient memory. Exiting...\n");
		  bzero((char *) t_lease, sizeof(*t_lease));
		  
		  t_lease->leased_to = req.client_link_local_addr;
		  
		  if(relay) {
		    t_lease->onlink                    = 0;
		    (t_lease->un).st.subnet_prefix     = req.relay_addr;
		    (t_lease->un).st.subnet_prefix_len = prefix_len;
		  }
		  else {
		    t_lease->onlink                    = 1;
		    strcpy((t_lease->un).interface_name, interface->name);
		  }

		  /* Imp: t_lease->scope                = s.scope; */
		  t_lease->xact_id              = req.transaction_id;
		  
		  if(s.L) {
		    t_lease->preferred_lifetime = preferred_lifetime;
		    t_lease->valid_lifetime     = valid_lifetime;
		  }
		  else {
		    t_lease->preferred_lifetime = 0;
		    t_lease->valid_lifetime     = 0;
		  }
		  strcpy(t_lease->dns_name, dns_name);
		  t_lease->abandoned          = 0;
		  t_lease->ip_addr            = ip_addr;
		  t_lease->ip_addr_prefix_len = s.prefix_len;

		  error = get_specific_addr(subnet, interface, t_lease, s.A, s.P, t_lease_list_head);
		  
		  if(error != REQUEST_GRANTED_NO_ERRORS) {
		    /* we can't continue further with this extension */
#ifdef DEBUG
		printf("get_specific_addr() couldn't grant requested addr\n");
#endif
		    if(status == 0)
		      status = Unavail;
		    memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			   sizeof(tl) + tl.length);
		    memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		    length_written += (sizeof(tl) + tl.length);
		    break;
		  }

		  if(s.Q && s.A) {
		    /* error will be first error encountered */
		    t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, A_RECORD);
		    if(error == 0)
		      error = t_error;
		    if(t_error == 0)
		      t_lease->A = 1;
		    else if(status == 0)
		      status = UnspecFail;
		  }
		  
		  if(s.Q && s.P) {
		    /* error will be first error encountered */
		    t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, P_RECORD);
		    if(error == 0)
		      error = t_error;
		    if(t_error == 0)
		      t_lease->P = 1;
		    else if(status == 0)
		      status = UnspecFail;
		  }
		  
		  if(!s.Q && (s.A || s.P)) {
		    /* Build a list of dns updates to be made later */
		    *temp_dns              = (DNS_UPDATE *) malloc(sizeof(DNS_UPDATE));
		    if(!*temp_dns)
		      errx(1, "Insufficient memory. Exiting...\n");
		    bzero((char *)*temp_dns, sizeof(**temp_dns));
		    
		    ((*temp_dns)->un).lease = t_lease;
		    (*temp_dns)->A          = s.A;
		    (*temp_dns)->P          = s.P;
		    (*temp_dns)->operation  = ADD;
		    (*temp_dns)->dhcp       = 1;
		    temp_dns                = &(*temp_dns)->next;
		  }
		  /* Write the extension with result in reply */
		  length_written = write_ip_address_extension(t_lease, length_written,
							      error, s.A, s.P);
		  *temp_lease = t_lease;
		  temp_lease  = &(*temp_lease)->next;
		  
		} // else for if(t_lease)
	      }// else if(s.C && !s.I)
	      else if(s.C && s.I) {
		/* The client has asked for DNS update of an addr obtained not thro'
		 * DHCP.
		 */
		uint8_t A_success = 0, P_success = 0;

		if(s.Q && s.A) {
		  /* error will be first error encountered */
		  t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, A_RECORD);

		  if(t_error == 0)
		    A_success = 1;
		  else if(status == 0)
		    status = UnspecFail;

		  if(error == 0)
		    error = t_error;
		}
		
		if(s.Q && s.P) {
		  /* error will be first error encountered */
		  t_error = update_dns(dns_name, &t_lease->ip_addr, ADD, P_RECORD);

		  if(t_error == 0)
		    P_success = 1;
		  else if(status == 0)
		    status = UnspecFail;

		  if(error == 0)
		    error = t_error;
		}
		
		if(!s.Q && (s.A || s.P)) {
		  /* Build a list of dns updates to be made later */
		  *temp_dns              = (DNS_UPDATE *) malloc(sizeof(DNS_UPDATE));
		  if(!*temp_dns)
		    errx(1, "Insufficient memory. Exiting...\n");
		  bzero((char *)*temp_dns, sizeof(**temp_dns));
		  
		  ((*temp_dns)->un).st.ip_addr = ip_addr;
		  strcpy(((*temp_dns)->un).st.dns_name, dns_name);
		  (*temp_dns)->A          = s.A;
		  (*temp_dns)->P          = s.P;
		  (*temp_dns)->operation  = ADD;
		  (*temp_dns)->dhcp       = 0;
		  temp_dns                = &(*temp_dns)->next;
		}
		/* Write the extension with result in reply */
		length_written = write_ip_address_extension_non_dhcp(length_written,
								     &ip_addr, error, A_success,
								     P_success, s.scope,
								     s.prefix_len);
	      }
	      
	    }/* else for if(c) */
	  }
	  break;


	case MAX_DHCP_MESSSAGE_SIZE_EXTENSION:
	  break;
	    
	case EXTENSION_REQUEST_EXTENSION:
	  {
	    uint16_t  temp_length = 0, old_length;
	    uint16_t  type;
	    HOST     *host;
	    TYPE_LEN  rep_tl;

	    /* temp_length stores amount of ERE I have read */

	    while(temp_length < tl.length) {
	      READ_DATA(&type, msg, length_read, sizeof(type));
	      temp_length += sizeof(type);
	      type = ntohs(type);
	      
	      switch(type) {

	      case POSIX_TIMEZONE_EXTENSION:
		break;

	      case DOMAIN_NAME_SERVER_EXTENSION:
		{
		  IN6_ADDR_WITH_PREFIX *temp_dns;

		  rep_tl.type    = htons(type);
		  rep_tl.length  = 0;

		  old_length = length_written;
		  //WRITE_DATA(send_msg, length_written, &rep_tl, sizeof(rep_tl));
		  length_written += sizeof(rep_tl);
		  if(relay) {
		    host = find_off_link_host(subnet, &req.client_link_local_addr);
		    if(host)
		      temp_dns = host->dn_server_list_head;
		    else
		      temp_dns = subnet->dn_server_list_head;
		  }
		  else {
		    host = find_on_link_host(interface, &req.client_link_local_addr);
		    if(host)
		      temp_dns = host->dn_server_list_head;
		    else
		      temp_dns = interface->dn_server_list_head;
		  }

		  for(; temp_dns; temp_dns = temp_dns->next) {
		    WRITE_DATA(send_msg, length_written, &temp_dns->prefix,
			       sizeof(temp_dns->prefix));
		    rep_tl.length += sizeof(temp_dns->prefix);
		  }

		  rep_tl.length = htons(rep_tl.length);
		  memcpy(send_msg + old_length, &rep_tl, sizeof(rep_tl));
		    
		}
		break;

	      case DOMAIN_NAME_SUFFIX_EXTENSION:
		{
		  char *result;

		  if(relay) {
		    host = find_off_link_host(subnet, &req.client_link_local_addr);
		    if(host)
		      result = host->domain_name_suffix;
		    else
		      result = subnet->domain_name_suffix;
		  }
		  else {
		    host = find_on_link_host(interface, &req.client_link_local_addr);
		    if(host)
		      result = host->domain_name_suffix;
		    else
		      result = interface->domain_name_suffix;
		  }
		  rep_tl.type   = htons(type);
		  rep_tl.length = htons(strlen(result));
		  WRITE_DATA(send_msg, length_written, &rep_tl, sizeof(rep_tl));
		  WRITE_DATA(send_msg, length_written, result, ntohs(rep_tl.length));

		}
		break;

	      default:
		break;
	      }
	    }
	  }
	  break;

	default:
	  break;
	}
	
      }
      
      if(r) {
	/* Imp: not yet implemented */
      }
      if(c) {
	/*
	 * Search the lease list to delete all leases except those present in the list
	 * we have built.
	 */
	temp_lease = &lease_list_head;
	while(*temp_lease) {
	  if(IN6_ARE_ADDR_EQUAL(&req.client_link_local_addr, &(*temp_lease)->leased_to)) {

	    if(relay) {
	      if(IN6_ARE_ADDR_EQUAL(&((*temp_lease)->un).st.subnet_prefix, &req.relay_addr)) {
		/* delete this lease if not present in the list we built */
		if(!is_present(*temp_lease, lie_head)) {
		  t_lease = *temp_lease;
		  *temp_lease = (*temp_lease)->next;

		  /* Undo DNS updates if done */
		  if(t_lease->A)
		    if(update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD) != 0
		       && status == 0)
		      status = UnspecFail;

		  if(t_lease->P)
		    if(update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, P_RECORD) != 0
		       && status == 0)
		      status = UnspecFail;

		  /* Imp: update lease file */
		  free(t_lease);
		  continue;
		}
	      }
	    }
	    else {
	      if(strcmp(((*temp_lease)->un).interface_name, interface->name) == 0) {
		/* delete this lease if not present in the list we built */
		if(!is_present(*temp_lease, lie_head)) {
		  t_lease     = *temp_lease;
		  *temp_lease = (*temp_lease)->next;

		  /* Undo DNS updates if done */
		  if(t_lease->A)
		    if(update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD) != 0
		       && status == 0)
		      status = UnspecFail;

		  if(t_lease->P)
		    if(update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, P_RECORD) != 0
		       && status == 0)
		      status = UnspecFail;

		  /* Imp: update lease file */
		  free(t_lease);
		  continue;
		}
	      }
	    }
	  }
	  temp_lease = &(*temp_lease)->next;
	}
	/* Write rep */
	SET_BITS_LOW(rep.r_status, status, 7);
	memcpy(send_msg, &rep, sizeof(rep));
	send_msg_to_client(send_msg, length_written, client, index);
      }
      else {
	TIME t;
	/* Write rep */
	SET_BITS_LOW(rep.r_status, status, 7);
	memcpy(send_msg, &rep, sizeof(rep));
	/* Update when leased for the t_lease_list and old leases which have been renewed */
	if(gettimeofday(&t, NULL) == -1)
	  errx(1, "Error in gettimeofday. Exiting...\n");

	for(t_lease = t_lease_list_head; t_lease; t_lease = t_lease->next) {
	  /* New leases */
	  t_lease->when_leased = t;
	}

	for(t_lie = lie_head; t_lie; t_lie = t_lie->next) {
	  /* Old leases */
	  t_lie->lease->when_leased = t;
	}

	/* Imp: Append the new leases and update old ones in stable storage */
	/* Imp: Send message. The extensions have been written before */
	send_msg_to_client(send_msg, length_written, client, index);

	/* Add the new leases to lease_list at the end */
	if(!lease_list_head)
	  lease_list_head = t_lease_list_head;
	else {
	  for(t_lease = lease_list_head; t_lease->next; t_lease = t_lease->next);
	  t_lease->next = t_lease_list_head;
	}

	/* update dns using dns list built up so far. */
	for(t_dns = dns_list_head; t_dns; t_dns = t_dns->next) {
	  if(t_dns->A) {
	    if(t_dns->dhcp)
	      t_dns->A_success = update_dns(((t_dns->un).lease)->dns_name,
					    &((t_dns->un).lease)->ip_addr, ADD, A_RECORD);
	    
	    else
	      t_dns->A_success = update_dns((t_dns->un).st.dns_name,
					    &(t_dns->un).st.ip_addr, ADD, A_RECORD);
	  }

	  if(t_dns->P) {
	    if(t_dns->dhcp)
	      t_dns->P_success = update_dns(((t_dns->un).lease)->dns_name,
					    &((t_dns->un).lease)->ip_addr, ADD, P_RECORD);
	    else
	      t_dns->P_success = update_dns((t_dns->un).st.dns_name,
					    &(t_dns->un).st.ip_addr, ADD, P_RECORD);
	  }
	  if(t_dns->dhcp) {
	    /* update_lease_file((t_dns->un).lease, t_dns->A_success, t_dns->P_success
	     * Note t_dns->A_success = 0 if update was successful
	     */
	  }
	  
	}
      }

      
      /* Now free up the list containing leases not to be deleted if c bit is set or
       * old leases if there are lease renewals
       */
      while(lie_head) {
	t_lie    = lie_head->next;
	free(lie_head);
	lie_head = t_lie;
      }
      /* Free up the dns list we have made */
      while(dns_list_head) {
	t_dns = dns_list_head->next;
	free(dns_list_head);
	dns_list_head = t_dns;
      }
    }
    break;

  case DHCP_RELEASE:
    {
      RELEASE rel;
      REPLY   rep;
      uint8_t r, status = 0, error, t_error;
      uint32_t rep_header_size;
      uint8_t ip_ext_present = 0;
      
      //LEASE_IN_EXT    *lie_head = NULL, **temp_lie = &lie_head, *t_lie;
      LEASE           **temp_lease, *t_lease;

      if(len < sizeof(rel)) {
	printf("Incorrect sized release message. Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      READ_DATA(&rel, msg, length_read, sizeof(rel));
      r = GET_BITS_HIGH(rel.r_reserved, 1, 8);
      rel.transaction_id = ntohs(rel.transaction_id);
      if(rel.transaction_id < 1024 || rel.transaction_id > 65535) {
	/* Note the second check rel.transaction_id > 65535 isn't reqd because of
	 * size of trensaction_id field.
	 */
#ifdef DEBUG
	printf("Transaction id specified in release is not in correct range."
	       " Discarding message...\n");
#endif
	break;/* break out of switch. You can use return as well */
      }

      if(!IN6_IS_ADDR_LINKLOCAL(&rel.client_link_local_addr)) {
	printf("Client's address specified in release message isn't link local."
	       " Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }
      
      if(!has_addr(&rel.server_addr, in_list_head)) {
	printf("Couldn't find server's address specified in release message."
	       " Discarding message\n");
	break;/* break out of switch. You can use return as well */
      }

      if(r) {
	/* x_addr is address of relay */
	if(IN6_IS_ADDR_UNSPECIFIED(&rel.x_addr)) {
#ifdef DEBUG
	  printf("Invalid relay address in client's message. Discarding...\n");
#endif
	  break;/* break out of switch. You can use return as well */
	}
	if(!check_scope(&rel.x_addr)) {
	  break;/* break out of switch. You can use return as well */
	}
	
      }

      rep.msg_type = DHCP_REPLY;
      rep.r_status = 0;
      SET_BITS_HIGH(rep.r_status, r, 1, 8);
      rep.transaction_id         = htons(rel.transaction_id);
      rep.client_link_local_addr = rel.client_link_local_addr;
      /*
       * Imp: I am not writing rep into send_msg but updating length_written 
       * so that I can attach extensions as required and later write rep. I
       * am not writing now because status field can't be filled in without
       * processing extensions
       */
      length_written += sizeof(rep);
      
      if(r)
	/* Add relay address to reply message */
	WRITE_DATA(send_msg, length_written, &rel.x_addr, sizeof(rel.x_addr));
      
      rep_header_size = length_written;

      client->sin6_port          = htons(CLIENT_PORT);
      /* Imp: Note what I am doing below */
      client->sin6_addr          = rel.x_addr;
      client->sin6_scope_id    = 0;
      if(r) {
	//client->sin6_addr        = rel.x_addr;
	//client->sin6_scope_id    = 0;
      }
      else {
	//client->sin6_addr        = rel.client_link_local_addr;
	//client->sin6_scope_id    = index;
      }
      
      while(length_read < len) {
	READ_DATA(&tl, msg, length_read, sizeof(tl));
	tl.type      = ntohs(tl.type);
	tl.length    = ntohs(tl.length);
	error        = 0;

	switch(tl.type) {
	  
	case IP_ADDRESS_EXTENSION:
	  {
	    struct{
	      uint32_t status     : 8;
	      uint32_t C          : 1;
	      uint32_t I          : 1;
	      uint32_t L          : 1;
	      uint32_t Q          : 1;
	      uint32_t A          : 1;
	      uint32_t P          : 1;
	      uint32_t reserved   : 8;
	      uint32_t scope      : 3;
	      uint32_t prefix_len : 7;
	    } s;
	    
	    uint32_t         temp_ext;
	    struct in6_addr  ip_addr;
	    uint32_t         preferred_lifetime, valid_lifetime, old_length, name_size;
	    char             dns_name[MAX_DNS_NAME];
	    
	    ip_ext_present = 1;
	    name_size      = 0;
	    old_length     = length_read;

	    READ_DATA(&temp_ext, msg, length_read, sizeof(temp_ext));
	    
	    temp_ext     = ntohl(temp_ext);
	    s.status     = GET_BITS_HIGH(temp_ext, 8, 32);
	    s.C          = GET_BITS_MIDDLE(temp_ext, 24, 1);
	    s.I          = GET_BITS_MIDDLE(temp_ext, 23, 1);
	    s.L          = GET_BITS_MIDDLE(temp_ext, 22, 1);
	    s.Q          = GET_BITS_MIDDLE(temp_ext, 21, 1);
	    s.A          = GET_BITS_MIDDLE(temp_ext, 20, 1);
	    s.P          = GET_BITS_MIDDLE(temp_ext, 19, 1);
	    s.reserved   = 0;
	    s.scope      = GET_BITS_MIDDLE(temp_ext, 10, 3);
	    s.prefix_len = GET_BITS_MIDDLE(temp_ext, 7, 7);
	    
	    if(s.C) {
	      READ_DATA(&ip_addr, msg, length_read, sizeof(struct in6_addr));
	    }
	    else {
	      /* Report error. Release without an address specified */
	      length_read = old_length + tl.length;
	      error = EXTENSION_SETTINGS_ILLEGAL;
	      if(status == 0)
		status = PoorlyFormed;
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;     
	    }
	    
	    if(s.I || s.Q) {
	      /* s.I => Can't release a non DHCP address */
	      /* s.Q => Only interpretation could be that client requires result of dns
	       * update(.., .., DELETE, ..) but nevertheless server replies after
	       * updating dns
	       */
	      length_read = old_length + tl.length;
	      error = EXTENSION_SETTINGS_ILLEGAL;
	      if(status == 0)
		status = PoorlyFormed;
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;
	    }
	    if(s.L) {
	      READ_DATA(& preferred_lifetime, msg, length_read, sizeof(preferred_lifetime));
	      READ_DATA(& valid_lifetime, msg, length_read, sizeof(valid_lifetime));
	      preferred_lifetime = ntohl(preferred_lifetime);
	      valid_lifetime     = ntohl(valid_lifetime);
	    }
	    
	    if((length_read - old_length) < tl.length) {
	      /* This means DNS name is present. Read it off */
	      name_size           = tl.length - (length_read - old_length);
	      READ_DATA(dns_name, msg, length_read, name_size);
	      dns_name[name_size] = '\0';
	      /* length_read = tl.length + old_length is also equally correct */
#ifdef DEBUG
	      printf("DNS name = %s in IP ADDRESS EXTENSION\n", dns_name);
#endif     
	    }
	    if((s.status == IP_ADDRESS_IN_USE) && (s.A || s.P)) {
	      /* Client has detected using DAD that IP address is in use. Then it must not
	       * ask to undo dns operations. It must ask for release of address.
	       */
	      error = EXTENSION_SETTINGS_ILLEGAL;
	      if(status == 0)
		status = PoorlyFormed;
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;
	    }
	    if(r) {
	      /* x_addr is addr of relay */
	      printf("r = 1, Finding %s leased to ",
		     inet_ntop(AF_INET6, &ip_addr, name, 64));
	      printf("%s\n", inet_ntop(AF_INET6, &rel.client_link_local_addr, name, 64));

	      temp_lease = find_lease_addr(&lease_list_head, &rel.client_link_local_addr,
				   &rel.x_addr, interface, &ip_addr);
	    }
	    else {
	      printf("r = 0, Finding %s leased to ",
		     inet_ntop(AF_INET6, &ip_addr, name, 64));
	      printf("%s\n", inet_ntop(AF_INET6, &rel.client_link_local_addr, name, 64));

	      temp_lease = find_lease_addr(&lease_list_head, &rel.client_link_local_addr,
				   NULL, interface, &ip_addr);
	    }
	    if(!temp_lease || !*temp_lease) {
	      printf("Client has asked to release %s. "
		     "But this address isn't present in server lease database\n",
		     inet_ntop(AF_INET6, &ip_addr, name, 64));
	      /* Report error: NoBinding */
	      error = REQUESTED_IP_ADDR_UNAVAILABLE;
	      if(status == 0)
		status = NoBinding;
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;

	    }
	    printf("(((((((((((((())))))))))))))found\n");
	    t_lease = *temp_lease;


	    if(s.status == IP_ADDRESS_IN_USE) {
	      /* Client has discovered through DAD that the IP address is
	       * already in use by another node.  The server MUST mark
	       * the errant IP address as unavailable for assignment, and
	       * SHOULD generate a log message indicating the problem to
	       * the administrator.
	       */

	      /* Imp: Do I need to undo dns updates       */
	      t_lease->abandoned = 1;
	      /* Imp: You have to define the function log_error. Currently I am using warnx */
	      warnx("Lease of ip addr %s must be disallowed becauese its in use by"
			" a different node\n",
			inet_ntop(AF_INET6, &t_lease->ip_addr, name, 64));

	      /* The server then generates a DHCP Reply
	       * message containing the client's IP address extension, with
	       * the ``status'' field set to 0 to indicate that it has
	       * received the client's release
	       */
	      error = 0;
	      /* Delete lease and write ext */
	      *temp_lease = (*temp_lease)->next;
	      free(t_lease);

	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		     sizeof(tl) + tl.length);
	      memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	      length_written += (sizeof(tl) + tl.length);
	      /* Continue processing further extensions */
	      break;
	      
	    }
	    if(s.A) {
	      /* delete DNS A records. Check if actually A record was updated.
	       * What todo if such a record wasn't updated at DNS?. We just give a warning.
	       */
	      if(t_lease->A == 1){
		t_error = update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD);

		if(error == 0)
		  error = t_error;

		if(t_error == 0)
		  t_lease->A = 0;
		else if(status == 0)
		  status = UnspecFail;
	      }
	      else {
#ifdef DEBUG
		warnx("Client is asking to undo dns update for A record for address %s."
		      " But the update hasn't been done. Ignoring the bug in message\n",
		      inet_ntop(AF_INET6, &ip_addr, name, 64));
#endif
	      }
	      /* Imp: update lease file */
	    }
	    if(s.P) {
	      /* delete DNS P records */
	      if(t_lease->P == 1) {
		t_error = update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, P_RECORD);

		if(error == 0)
		  error = t_error;

		if(t_error == 0)
		  t_lease->A = 0;
		else if(status == 0)
		  status = UnspecFail;
	      }
	      else {
#ifdef DEBUG
		warnx("Client is asking to undo dns update for P record for address %s."
		      " But the update hasn't been done. Ignoring the bug in message\n",
		      inet_ntop(AF_INET6, &ip_addr, name, 64));
#endif
		
	      }
	      /* Imp: update lease file */
	    }
	    if(!s.A && !s.P) {
	      /* Client is asking to release the address. Note undo DNS updates. */
	      
	      /* Build up a list of addresses to release and release later when
	       * you are sure that no extension contains an addr which is
	       * same as x_addr if r == 0 because then client is releasing an address
	       * on which it wants a reply back.
	       */
	      /* Just now
	      *temp_lie          = (LEASE_IM_EXT *)malloc(sizeof(LEASE_IM_EXT));
	      (*temp_lie)->lease = t_lease;
	      temp_lie           = &(*temp_lie)->next;
	      */
	      if(r == 0 && IN6_ARE_ADDR_EQUAL(&rel.x_addr, &t_lease->ip_addr)) {
#ifdef DEBUG
		printf("Client wishes to release %s and wants reply back on the same addr.\n",
		       inet_ntop(AF_INET6, &t_lease->ip_addr, name, 64));
#endif		
		/* Send error back */
		error = REQUESTED_IP_ADDR_UNAVAILABLE;
		if(status == 0)
		  status = PoorlyFormed;
		memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		       sizeof(tl) + tl.length);
		memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
		length_written += (sizeof(tl) + tl.length);
		/* Continue processing further extensions */
		break;
		
	      }

	      if(t_lease->A) {
		t_error = update_dns(t_lease->dns_name, &t_lease->ip_addr,
				     DELETE, A_RECORD);

		if(error == 0)
		  error = t_error;

		if(t_error == 0)
		  t_lease->A = 0;/* We needn't do this as in a moment lease
				    will be deleted */
		else if(status == 0)
		  status = UnspecFail;
	      }
	      if(t_lease->P) {
		t_error =  update_dns(t_lease->dns_name, &t_lease->ip_addr,
				      DELETE, P_RECORD);           

		if(error == 0)
		  error = t_error;
		
		if(t_error == 0)
		  t_lease->A = 0;/* We needn't do this as in a moment lease
				    will be deleted */
		else if(status == 0)
		  status = UnspecFail;
	      }
	      /* Just now
	      t_lease->status_addr = length_written + sizeof(tl);
	      memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
			 sizeof(tl) + tl.length);
	      //memcpy(send_msg + length_written + sizeof(tl), &s.status,
	      sizeof(s.status));
	      length_written += sizeof(tl) + tl.length;
	      */
	      /* Imp: Delete lease and update lease file */
	      *temp_lease = (*temp_lease)->next;
	      free(t_lease);
	    }
	    /* Write result: extension */
	    memcpy(send_msg + length_written, msg + old_length - sizeof(tl),
		   sizeof(tl) + tl.length);
	    memcpy(send_msg + length_written + sizeof(tl), &error, sizeof(error));
	    length_written += (sizeof(tl) + tl.length);
	  }
	  break;
	  
	default:
	  /* Only IP addr extension is allowed in a release message
	   * Report error and ignore extension.
	   */
	  if(status == 0)
	    status = PoorlyFormed;
	  length_read += tl.length;
	  break;
	}
      }/* while(length_read < len) */
      /* Now release all the specified resources */

      if(!ip_ext_present) {
	/* If no extensions are included in the Release message, then all releasable
	 * resources associated with the client's binding are to be released.
	 */
	temp_lease = &lease_list_head;
	while(*temp_lease) {
	  if(IN6_ARE_ADDR_EQUAL(&rel.client_link_local_addr, &(*temp_lease)->leased_to)) {
	    
	    if(r) {
	      if(IN6_ARE_ADDR_EQUAL(&((*temp_lease)->un).st.subnet_prefix, &rel.x_addr)) {
		/* delete this lease from the list we built */
		t_lease = *temp_lease;
		*temp_lease = (*temp_lease)->next;

		/* Undo DNS updates if done */
		if(t_lease->A) {
		  t_error = update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD);
		  
		  if(t_error == 0)
		    t_lease->A = 0;/* We needn't do this as in a moment lease will be deleted */
		  else if(status == 0)
		    status = UnspecFail;
		}
		if(t_lease->P) {
		  t_error =  update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, P_RECORD);
		  
		  if(t_error == 0)
		    t_lease->A = 0;/* We needn't do this as in a moment lease will be deleted */
		  else if(status == 0)
		    status = UnspecFail;
		}
		/* Imp: update lease file */
		free(t_lease);
		continue;
	      }
	    }
	    else {
	      if(strcmp(((*temp_lease)->un).interface_name, interface->name) == 0) {
		t_lease = *temp_lease;
		*temp_lease = (*temp_lease)->next;
		
		/* Undo DNS updates if done */
		if(t_lease->A) {
		  t_error = update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, A_RECORD);
		  
		  if(t_error == 0)
		    t_lease->A = 0;/* We needn't do this as in a moment lease will be deleted */
		  else if(status == 0)
		    status = UnspecFail;
		}
		if(t_lease->P) {
		  t_error =  update_dns(t_lease->dns_name, &t_lease->ip_addr, DELETE, P_RECORD);
		  
		  if(t_error == 0)
		    t_lease->A = 0;/* We needn't do this as in a moment lease will be deleted */
		  else if(status == 0)
		    status = UnspecFail;
		}
		/* Imp: update lease file */
		free(t_lease);
		continue;
	      }
	    }
	  }
	  temp_lease = &(*temp_lease)->next;
	}/* while(*temp_lease) */

	/* Just now
	SET_BITS_LOW(rep.r_status, Success, 7);
	send_msg_to_client(send_msg, length_written, client, index);
	return;
	*/
      }

      /* Send reply */
      SET_BITS_LOW(rep.r_status, status, 7);
      memcpy(send_msg, &rep, sizeof(rep));
      send_msg_to_client(send_msg, length_written, client, index);

      /* Now free up the list we have made */
      /*
      while(lie_head) {
	t_lie    = lie_head->next;
	free(lie_head);
	lie_head = t_lie;
	}*/
      
    }
    break;
    
  case DHCP_RECONFIGURE_REPLY:
    {

    }
    break;

  default:
#ifdef DEBUG
    printf("Message of unknown DHCP type received. Discarding...\n");
#endif
    break;
  }
}
