# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>
# include "int.h"
# include "max_dns_name.h"
# include "in6_addr_list.h"
# include "lease_structure.h"

# define MAX_TOKENS_PER_LINE 8
# define MAX_TOKEN_LEN       256

/************************************************/
typedef struct timeval TIME;
/************************************************/

char  *LEASE_CONF_FILE;
char   line[MAX_TOKENS_PER_LINE][MAX_TOKEN_LEN];
uint   line_no = 0;
char   name[64];
/*
int read_line(FILE *fp){
  int i = 0, n;
  char c;

  while(1){
    fscanf(fp, "%*[ \t]");
    n = fscanf(fp, "%[^\n\t ]", line[i]);
    if(n <= 0){
      c = fgetc(fp);
      if(c == '\n'){
	line_no++;
	break;
      }
      if(c == EOF)
	return -1;
    }
    i++;
  }
  return i;
}

void print_line(int i){
  int k;

  //printf("%d\n", i);
  for(k = 0; k < i; k++)
    printf("%s ", line[k]);
  printf("\n");
}

int read_useful_line(FILE *fp){
  int i;

  while(1){
    i = read_line(fp);
    //printf(".................................i = %d\n", i);
    //print_line(i);
    if(i == -1)
      return i;
    if(i != 0 && line[0][0] != '#')
      return i;
  }
}

uint8_t get_and_delete_prefix_len(char addr[]){
  int i;

  for(i = strlen(addr) - 1; i > 0; i--)
    if(addr[i] == '/')
      break;

  if(i == 0)
    errx(1, "Address doesn't contain prefix. Successfully read %u lines\n", line_no);

  addr[i] = '\0';
  return atoi(addr + i + 1);
}
*/
LEASE *read_lease(FILE *fp){
  int                 i;
  LEASE              *lease;
  IN6_ADDR_LIST     **t;
  
  i = read_useful_line(fp);
  switch(i){

  case -1:
    /* EOF reached */
    return NULL;
    break;
    
  case 4:
    if(line[1][strlen(line[1]) - 1] != '{')
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   LEASE_CONF_FILE, line_no);

    line[1][strlen(line[1]) - 1] = '\0';
    break;
    
  case 5:
    break;
    
  default:
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 LEASE_CONF_FILE, line_no);
    break;
  }

  if(strcmp(line[0], "leased-to") != 0)
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 LEASE_CONF_FILE, line_no);
  
  lease = (LEASE *) malloc(sizeof(LEASE));
  if(!lease)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) lease, sizeof(LEASE));
  //lease->A = lease->P = 2;/* A value not possible for A or P */
  lease->dns_name[0] = '\0';
  if(inet_pton(AF_INET6, line[1], &lease->leased_to) <= 0)
    errx(1, "Error in converting link local addr to binary\n");

  if(strcmp(line[2], "subnet-prefix") == 0){
    lease->onlink = 0;
    (lease->un).st.subnet_prefix_len = get_and_delete_prefix_len(line[3]);
    if(inet_pton(AF_INET6, line[3], &(lease->un).st.subnet_prefix) <= 0)
      errx(1, "Error in converting subnet-prefix to binary\n");
  }
  else if(strcmp(line[2], "interface") == 0){
    lease->onlink = 1;
    strcpy((lease->un).interface_name, line[3]);
  }
  else
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 LEASE_CONF_FILE, line_no);
  
  while(1){
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      //return in;
      errx(2, "Unexpected end of file. Successfully read %u lines\n", line_no);
      break;
      
    case 1:
      if(strcmp(line[0], "}") == 0){
	//printf("End of lease\n\n");
	return lease;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     LEASE_CONF_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "ip-address") == 0){
	
	if(!IN6_IS_ADDR_UNSPECIFIED(&lease->ip_addr)){
	  warnx("ip-address mentioned more than once. New address will be used\n");
	}
	
        lease->ip_addr_prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &lease->ip_addr) <= 0)
	  errx(1, "Error in converting ip addr to binary\n");
	
      }
      else if(strcmp(line[0], "scope") == 0){
	lease->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "transaction-id") == 0){
        /* Remember transaction id used by client is in range 1024-65535 */
	if(lease->xact_id != 0)
	  warnx("transaction-id specified more than once\n");
	lease->xact_id = atoi(line[1]);
      }
      else if(strcmp(line[0], "preferred-lifetime") == 0){
	if(lease->preferred_lifetime > 0)
	  warnx("preferred lifetime mentioned more than once. Will accept new one\n");

	lease->preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "valid-lifetime") == 0){
	if(lease->valid_lifetime > 0)
	  warnx("valid lifetime mentioned more than once. Will accept new one\n");

	lease->valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-name") == 0){
	if(lease->dns_name[0] != '\0')
	  warnx("dns-name mentioned more than once. Will accept new one\n");
	strcpy(lease->dns_name, line[1]);
      }
      else if(strcmp(line[0], "default-preferred-lifetime") == 0){
	lease->default_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	lease->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
	if(lease->max_preferred_lifetime > 0)
	  warnx("max preferred lifetime mentioned more than once. Will accept new one\n");

	lease->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	if(lease->max_valid_lifetime > 0)
	  warnx("max valid lifetime mentioned more than once. Will accept new one\n");

	lease->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "A") == 0){
	lease->A = atoi(line[1]);
      }
      else if(strcmp(line[0], "P") == 0){
	lease->P = atoi(line[1]);
      }
      else{
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     LEASE_CONF_FILE, line_no);
      }
       
      break;

    case 3:
      if(strcmp(line[0], "when-leased") == 0){
         if(lease->when_leased.tv_sec != 0 || lease->when_leased.tv_usec != 0)
            errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	         LEASE_CONF_FILE, line_no);
	 lease->when_leased.tv_sec  = atol(line[1]);
         lease->when_leased.tv_usec = atol(line[2]);
      }
      else
         errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	      LEASE_CONF_FILE, line_no);

      break;

    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   LEASE_CONF_FILE, line_no);
      break;
    }
  }
}

void print_lease(LEASE *lease){
  printf("leased-to: %s\n",
	 inet_ntop(AF_INET6, &lease->leased_to, name, 64));
  if(lease->onlink){
    printf("On interface: %s\n", (lease->un).interface_name);
  }
  else{
    printf("On subnet: %s/%u\n",
	   inet_ntop(AF_INET6, &(lease->un).st.subnet_prefix, name, 64),
	   (lease->un).st.subnet_prefix_len);
  }
  printf("ip_address: %s/%u\n",
	 inet_ntop(AF_INET6, &lease->ip_addr, name, 64),
	 lease->ip_addr_prefix_len);
  printf("scope: %u\n", lease->scope);
  printf("transaction_id: %u\n", lease->xact_id);
  printf("when_leased: %u %u\n", lease->when_leased.tv_sec, lease->when_leased.tv_usec);
  printf("preferred_lifetime: %u\n", lease->preferred_lifetime);
  printf("valid_lifetime: %u\n", lease->valid_lifetime);
  printf("default-preferred_lifetime: %u\n", lease->default_preferred_lifetime);
  printf("default-valid_lifetime: %u\n", lease->default_valid_lifetime);
  printf("max-preferred_lifetime: %u\n", lease->max_preferred_lifetime);
  printf("max-valid_lifetime: %u\n", lease->max_valid_lifetime);
  printf("A: %u\n", lease->A);
  printf("P: %u\n", lease->P);
  printf("dns-name: %s\n", lease->dns_name);
  printf("\n\n");

}

void read_lease_file(char *file, LEASE **lease){
  FILE *fp;
  int i, k;
  //LEASE **lease = &lease_list_head;

  LEASE_CONF_FILE = file;
  fp = fopen(LEASE_CONF_FILE, "r");
  if(!fp)
    errx(1, "Error opening %s. Exiting...\n", LEASE_CONF_FILE);

  while(1){

    *lease = read_lease(fp);
    if(!(*lease))
      break;
    lease = &(*lease)->next;
    
  }
}
/*
main(int argc, char *argv[]){
  LEASE *lease;
  LEASE *lease_list_head;

  if(argc != 2){
    printf("Incorrect number of arguments to main\n");
    exit(0);
  }

  read_lease_file(argv[1], &lease_list_head);
  for(lease = lease_list_head; lease; lease = lease->next)
    print_lease(lease);
}
*/
