# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>
# include "../int.h"
# define MAX_TOKENS_PER_LINE 8
# define MAX_TOKEN_LEN       256
# define DEBUG

typedef struct range{

  struct range    *next;
  struct subnet   *subnet;
  struct in6_addr  start;
  struct in6_addr  end;
  uint32_t         default_preferred_lifetime;
  uint32_t         default_valid_lifetime;
  uint32_t         max_preferred_lifetime;
  uint32_t         max_valid_lifetime;
  uint8_t          scope: 3;
  uint8_t          dns_update_A_allowed: 1;
  uint8_t          dns_update_P_allowed: 1;

}RANGE;

typedef struct in6_addr_with_prefix{

  struct in6_addr_with_prefix *next;
  struct in6_addr              prefix;
  uint8_t                      prefix_len;

}IN6_ADDR_WITH_PREFIX;

typedef struct in6_addr_list{

  struct in6_addr_list *next;
  struct in6_addr       addr;

}IN6_ADDR_LIST;

typedef IN6_ADDR_WITH_PREFIX PREFIX_LIST;

typedef struct host{
  struct host      *next;
  struct subnet    *subnet;
  struct in6_addr   link_local_addr;
  struct in6_addr   subnet_prefix;
  uint8_t           subnet_prefix_len;

  uint32_t          default_preferred_lifetime;
  uint32_t          default_valid_lifetime;
  uint32_t          max_preferred_lifetime;
  uint32_t          max_valid_lifetime;
  uint8_t           scope: 3;
  uint8_t           dns_update_A_allowed: 1;
  uint8_t           dns_update_P_allowed: 1;
  uint8_t           preference;

  IN6_ADDR_LIST    *fixed_addr_list_head;
}HOST;

typedef struct subnet{
  IN6_ADDR_WITH_PREFIX *addr_list_head;

  uint32_t              default_preferred_lifetime;
  uint32_t              default_valid_lifetime;
  uint32_t              max_preferred_lifetime;
  uint32_t              max_valid_lifetime;
  uint8_t               scope: 3;
  uint8_t               dns_update_A_allowed: 1;
  uint8_t               dns_update_P_allowed: 1;
  uint8_t               preference;

  PREFIX_LIST          *adv_prefix_list_head;
  IN6_ADDR_WITH_PREFIX  server_addr;
  HOST                 *host_list_head;
  RANGE                *range_list_head;
  struct subnet        *next;
}SUBNET;

char *SUBNET_PROPERTIES_FILE;
char line[MAX_TOKENS_PER_LINE][MAX_TOKEN_LEN];
uint line_no = 0;
char name[64];
SUBNET *subnet_list_head;//, *sub_list_head;

int read_line(FILE *fp){
  int i = 0, n, k = 0;
  char c;
  
  while(1){
    fscanf(fp, "%*[ \t]");
#ifdef DEBUG
    printf("\n......b: %d\n", subnet_list_head);
#endif

    n = fscanf(fp, "%[^\n\t ]", line[i]);

#ifdef DEBUG
    printf("_______a: %d\n", subnet_list_head);
#endif
#ifdef DEBUG
    //getchar();
#endif
    if(n <= 0){
      c = fgetc(fp);
      if(c == '\n'){
	line_no++;
	break;
      }
      if(c == EOF)
	return -1;
    }
    i++;
  }
  return i;
}


void print_line(int i){
  int k;
  
  printf("%d: ", i);
  for(k = 0; k < i; k++)
    printf("<%s> ", line[k]);
  printf("\n");
}

int read_useful_line(FILE *fp){
  int i;

  while(1){
    i = read_line(fp);
    if(i == -1)
      return i;
    if(i != 0 && line[0][0] != '#')
      return i;
  }
}

uint8_t get_and_delete_prefix_len(char addr[]){
  int i;
  
  for(i = strlen(addr) - 1; i > 0; i--)
    if(addr[i] == '/')
      break;
  
  if(i == 0)
    errx(1, "Address doesn't contain prefix. Successfully read %u lines\n", line_no);
  
  addr[i] = '\0';
  return atoi(addr + i + 1);
}

RANGE *read_range(FILE *fp, RANGE **r, SUBNET *subnet){
  int    i;
  RANGE *range;
 
  while(*r)
    r = &(*r)->next;

  range = (RANGE *) malloc(sizeof(RANGE));
  if(!range)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) range, sizeof(RANGE));
  if(inet_pton(AF_INET6, line[1], &range->start) <= 0)
    errx(1, "Error in converting range start address to binary\n");
  
  if(inet_pton(AF_INET6, line[2], &range->end) <= 0)
    errx(1, "Error in converting range end address to binary\n");
  
  /* Set the values to upper level default */
  range->subnet                     = subnet;
  range->default_preferred_lifetime = subnet->default_preferred_lifetime;
  range->default_valid_lifetime     = subnet->default_valid_lifetime;
  range->max_preferred_lifetime     = subnet->max_preferred_lifetime;
  range->max_valid_lifetime         = subnet->max_valid_lifetime;
  range->scope                      = subnet->scope;
  range->dns_update_A_allowed       = subnet->dns_update_A_allowed;
  range->dns_update_P_allowed       = subnet->dns_update_P_allowed;
  //range->preference                 = subnet->preference;

  while(1){
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      errx(1, "Unexpected end of file\n");
      break;
      
    case 1:
      if(strcmp(line[0], "}") == 0){
	*r = range;
	return;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	range->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	range->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        range->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	range->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	range->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	range->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	range->dns_update_P_allowed = atoi(line[1]);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   SUBNET_PROPERTIES_FILE, line_no);
      break;
    }
  }
}

void read_prefixes(FILE *fp, PREFIX_LIST **p){
  PREFIX_LIST *pre;
  int          i, k;

  while(*p)
    p = &(*p)->next;

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "}") == 0)
	return;
      else{
	pre = (PREFIX_LIST *)malloc(sizeof(PREFIX_LIST));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(PREFIX_LIST));
	pre->prefix_len = get_and_delete_prefix_len(line[0]);
	if(inet_pton(AF_INET6, line[0], &pre->prefix) <= 0)
	  errx(1, "Error in converting subnet prefix to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    
    default:
      for(k = 0; k < i; k++){
	pre = (PREFIX_LIST *)malloc(sizeof(PREFIX_LIST));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(PREFIX_LIST));
	pre->prefix_len = get_and_delete_prefix_len(line[k]);
	if(inet_pton(AF_INET6, line[k], &pre->prefix) <= 0)
	  errx(1, "Error in converting subnet prefix to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    }
  }
}

void read_host(FILE *fp, HOST **h, SUBNET *subnet){
  HOST           *host;
  int             i, k;
  IN6_ADDR_LIST **t;

  host = (HOST *)malloc(sizeof(HOST));
  if(!host)
    errx(1, "Insufficient memory\n");
  bzero((char *)host, sizeof(HOST));
  
  if(inet_pton(AF_INET6, line[1], &host->link_local_addr) <= 0)
    errx(1, "Error in converting host's link-local-address to binary\n");

  host->subnet_prefix_len = get_and_delete_prefix_len(line[3]);
  if(inet_pton(AF_INET6, line[3], &host->subnet_prefix) <= 0)
    errx(1, "Error in converting host's subnet-prefix to binary\n");

  /* Set the values to upper level default */
  host->subnet                     = subnet;
  host->default_preferred_lifetime = subnet->default_preferred_lifetime;
  host->default_valid_lifetime     = subnet->default_valid_lifetime;
  host->max_preferred_lifetime     = subnet->max_preferred_lifetime;
  host->max_valid_lifetime         = subnet->max_valid_lifetime;
  host->scope                      = subnet->scope;
  host->dns_update_A_allowed       = subnet->dns_update_A_allowed;
  host->dns_update_P_allowed       = subnet->dns_update_P_allowed;
  host->preference                 = subnet->preference;
  
  t = &host->fixed_addr_list_head;

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "}") == 0){
	while(*h)
	  h = &(*h)->next;
	*h = host;
	return;
      }
      break;
    
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	host->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	host->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        host->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	host->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	host->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	host->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	host->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "preference") == 0){
	host->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "fixed-address") == 0){
	*t = (IN6_ADDR_LIST *)malloc(sizeof(IN6_ADDR_LIST));
	if(!(*t))
	  errx(1, "Insufficient memory\n");
	bzero((char *)(*t), sizeof(IN6_ADDR_LIST));
	if(inet_pton(AF_INET6, line[1], &(*t)->addr) <= 0)
	  errx(1, "Error in converting host's fixed address to binary\n");
	
	t = &(*t)->next;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;

    default:
      if(strcmp(line[0], "fixed-address") == 0){
	for(k = 1; k < i; k++){
	  *t = (IN6_ADDR_LIST *)malloc(sizeof(IN6_ADDR_LIST));
	  if(!(*t))
	    errx(1, "Insufficient memory\n");
	  bzero((char *)(*t), sizeof(IN6_ADDR_LIST));
	  if(inet_pton(AF_INET6, line[k], &(*t)->addr) <= 0)
	    errx(1, "Error in converting host's fixed address to binary\n");
	  
	  t = &(*t)->next;
	}
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
    }
  }
}

SUBNET *read_subnet(FILE *fp){
  int         i, enclosing_block_ends, k;
  SUBNET     *subnet;
  RANGE     **r;
  IN6_ADDR_WITH_PREFIX **t;
  
  i = read_useful_line(fp);
  switch(i){
    
  case -1:
    /* EOF reached */
    return NULL;
    break;
    
  case 1:
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 SUBNET_PROPERTIES_FILE, line_no);
    break;

  case 2:
    if(line[1][strlen(line[1]) - 1] != '{')
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   SUBNET_PROPERTIES_FILE, line_no);
    
    line[1][strlen(line[1]) - 1] = '\0';
    break;
      
  default:
    break;
  }
  i--;
  if(strcmp(line[i], "{") == 0)
    i--;
  else if(line[i][strlen(line[i]) - 1] == '{')
    line[i][strlen(line[i]) - 1] = '\0';
  else
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 SUBNET_PROPERTIES_FILE, line_no);

  if(strcmp(line[0], "subnet") != 0)
    errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	 SUBNET_PROPERTIES_FILE, line_no);
  
  subnet = (SUBNET *) malloc(sizeof(SUBNET));
  if(!subnet)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) subnet, sizeof(SUBNET));
  /* Default values for dns update of A and P records */
  subnet->dns_update_A_allowed = subnet->dns_update_A_allowed = 1;
  t = &subnet->addr_list_head;

  for(k = 1; k <= i; k++){
    *t = (IN6_ADDR_WITH_PREFIX *)malloc(sizeof(IN6_ADDR_WITH_PREFIX));
    if(!(*t))
      errx(1, "Insufficient memory\n");
    bzero((char *)(*t), sizeof(IN6_ADDR_WITH_PREFIX));
    (*t)->prefix_len = get_and_delete_prefix_len(line[k]);
    if(inet_pton(AF_INET6, line[k], &(*t)->prefix) <= 0)
      errx(1, "Error in converting subnet prefix to binary\n");
    t = &(*t)->next;
  }

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "list-of-prefixes-to-advertise{") == 0){
	read_prefixes(fp, &subnet->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "}") == 0){
	return subnet;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	subnet->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	subnet->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        subnet->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	subnet->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	subnet->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	subnet->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	subnet->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "preference") == 0){
	subnet->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "list-of-prefixes-to-advertise") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_prefixes(fp, &subnet->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "server-address") == 0){
	(subnet->server_addr).prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(subnet->server_addr).prefix) <= 0)
	  errx(1, "Error in converting server address to binary\n");
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    case 3:
      if(strcmp(line[0], "range") == 0 && line[2][strlen(line[2]) - 1] == '{'){
	line[2][strlen(line[2]) - 1] = '\0';
	/* Note: First goto end of list of range and then add */
	read_range(fp, &subnet->range_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    case 4:
      if(strcmp(line[0], "range") == 0 && strcmp(line[3], "{") == 0){
	/* Note: First goto end of list of range and then add */
	read_range(fp, &subnet->range_list_head, subnet);
      }
      else if(strcmp(line[0], "host") == 0 &&
	      line[3][strlen(line[3]) - 1] == '{' &&
	      strcmp(line[2], "subnet-prefix") == 0){
	line[3][strlen(line[3]) - 1] = '\0';
	read_host(fp, &subnet->host_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;

    case 5:
      if(strcmp(line[0], "host") == 0 &&
	 strcmp(line[4], "{") == 0 &&
	 strcmp(line[2], "subnet-prefix") == 0){
	read_host(fp, &subnet->host_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   SUBNET_PROPERTIES_FILE, line_no);
      break;
      
    }
  }
  warnx("This should be impossible\n");
  return subnet;
}

void print_subnet(SUBNET *subnet){
  IN6_ADDR_WITH_PREFIX *temp_addr_pre;
  IN6_ADDR_LIST        *temp_addr;
  PREFIX_LIST          *temp_pre;
  RANGE                *temp_range;
  HOST                 *temp_host;

  printf("subnet: ");
  for(temp_addr_pre = subnet->addr_list_head; temp_addr_pre;
      temp_addr_pre = temp_addr_pre->next){
    printf("%s/%u ",inet_ntop(AF_INET6, &temp_addr_pre->prefix, name, 64),
	   temp_addr_pre->prefix_len);
  }
  printf("\n");
  /* Print top level variables */
  printf("default-preferred-lifetime: %u\n", subnet->default_preferred_lifetime);
  printf("default-valid-lifetime: %u\n", subnet->default_valid_lifetime);
  printf("max-preferred-lifetime: %u\n", subnet->max_preferred_lifetime);
  printf("max-valid-lifetime: %u\n", subnet->max_valid_lifetime);
  printf("scope: %u\n", subnet->scope);
  printf("dns-update-A-allowed: %u\n", subnet->dns_update_A_allowed);
  printf("dns-update-P-allowed: %u\n", subnet->dns_update_P_allowed);
  printf("preference: %u\n\n", subnet->preference);

  printf("list-of-prefixes-to-advertise: ");
  for(temp_pre = subnet->adv_prefix_list_head; temp_pre; temp_pre = temp_pre->next){
    printf("%s/%u ", inet_ntop(AF_INET6, &temp_pre->prefix, name, 64), temp_pre->prefix_len);
  }
  printf("\n\n");
  printf("server-address: %s/%u\n",
	 inet_ntop(AF_INET6, &(subnet->server_addr).prefix, name, 64),
	 (subnet->server_addr).prefix_len);

  /* Print host declarations */
  for(temp_host = subnet->host_list_head; temp_host; temp_host = temp_host->next){
    printf("\nhost %s ", inet_ntop(AF_INET6, &temp_host->link_local_addr, name, 64));
    printf("subnet-prefix %s/%u\n",
	   inet_ntop(AF_INET6, &temp_host->subnet_prefix, name, 64),
	   temp_host->subnet_prefix_len);

    printf("default-preferred-lifetime: %u\n", temp_host->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n", temp_host->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n", temp_host->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n", temp_host->max_valid_lifetime);
    printf("scope: %u\n", temp_host->scope);
    printf("dns-update-A-allowed: %u\n", temp_host->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n", temp_host->dns_update_P_allowed);
    printf("preference: %u\n\n", temp_host->preference);

    printf("fixed address ");
    for(temp_addr = temp_host->fixed_addr_list_head; temp_addr; temp_addr = temp_addr->next){
      printf("%s ", inet_ntop(AF_INET6, &temp_addr->addr, name, 64));
      
    }
    printf("\n\n");
  }
  /* Print range declarations */
  for(temp_range = subnet->range_list_head; temp_range; temp_range = temp_range->next){
    printf("________________________________________________\nrange %s ",
	   inet_ntop(AF_INET6, &temp_range->start, name, 64));
    printf("%s\n", inet_ntop(AF_INET6, &temp_range->end, name, 64));
    printf("default-preferred-lifetime: %u\n", temp_range->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n", temp_range->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n", temp_range->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n", temp_range->max_valid_lifetime);
    printf("scope: %u\n", temp_range->scope);
    printf("dns-update-A-allowed: %u\n", temp_range->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\n", temp_range->dns_update_P_allowed);
  }
}

void parse_subnet_properties_file(){
  FILE *fp;
  int i, k;
  SUBNET **subnet = &subnet_list_head;
  
  fp = fopen(SUBNET_PROPERTIES_FILE, "r");
  if(!fp)
    errx(1, "Error opening %s. Exiting...\n", SUBNET_PROPERTIES_FILE);
  
  while(1){
    *subnet = read_subnet(fp);
    if(!(*subnet))
      break;
    subnet = &(*subnet)->next;
    
  }
}

main(int argc, char *argv[]){
  SUBNET *sub;

  if(argc != 2){
    printf("Incorrect number of arguments to main\n");
    exit(0);
  }
  
  SUBNET_PROPERTIES_FILE = argv[1];
  parse_subnet_properties_file();
  for(sub = subnet_list_head; sub; sub = sub->next){
    print_subnet(sub);
  }
}


