# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>

# define MAX_TOKENS_PER_LINE 8
# define MAX_TOKEN_LEN       256

# define FIXED_ADDRESS_LEVEL 2
# define MAX_RANGE_LEVEL     2

# include "max_dns_name.h"
# include "int.h"
# include "in6_addr_list.h"
# include "host_subnet_interface_structures.h"

# define IN6_IS_ADDR_GLOBAL(x) ( !IN6_IS_ADDR_UNSPECIFIED(x) && !IN6_IS_ADDR_LOOPBACK(x) && \
!IN6_IS_ADDR_MULTICAST(x) && !IN6_IS_ADDR_LINKLOCAL(x) && !IN6_IS_ADDR_SITELOCAL(x))

char *ADDRESS_PROPERTIES_FILE;
char line[MAX_TOKENS_PER_LINE][MAX_TOKEN_LEN];
uint line_no;
char name[64];

uint8_t get_and_delete_scope(char addr[]){
  int i;
  
  for(i = strlen(addr) - 1; i > 0; i--)
    if(addr[i] == '-')
      break;
  
  if(i == 0)
    return 0;
  
  addr[i] = '\0';
  return atoi(addr + i + 1);
}

void read_domain_name_server(FILE *fp, IN6_ADDR_WITH_PREFIX **p){
  IN6_ADDR_WITH_PREFIX *pre;
  int          i, k;

  while(*p)
    p = &(*p)->next;

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "}") == 0) {
	return;
      }
      else {
	pre = (IN6_ADDR_WITH_PREFIX *)malloc(sizeof(IN6_ADDR_WITH_PREFIX));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(IN6_ADDR_WITH_PREFIX));
	pre->prefix_len = get_and_delete_prefix_len(line[0]);
	if(inet_pton(AF_INET6, line[0], &pre->prefix) <= 0)
	  errx(1, "Error in converting dns-address to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    
    default:
      for(k = 0; k < i; k++){
	pre = (IN6_ADDR_WITH_PREFIX *)malloc(sizeof(IN6_ADDR_WITH_PREFIX));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(IN6_ADDR_WITH_PREFIX));
	pre->prefix_len = get_and_delete_prefix_len(line[k]);
	if(inet_pton(AF_INET6, line[k], &pre->prefix) <= 0)
	  errx(1, "Error in converting dns-address to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    }
  }
}

void read_prefixes(FILE *fp, PREFIX_LIST **p){
  PREFIX_LIST *pre;
  int          i, k;

  while(*p)
    p = &(*p)->next;

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "}") == 0)
	return;
      else{
	pre = (PREFIX_LIST *)malloc(sizeof(PREFIX_LIST));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(PREFIX_LIST));
	pre->prefix_len = get_and_delete_prefix_len(line[0]);
	if(inet_pton(AF_INET6, line[0], &pre->prefix) <= 0)
	  errx(1, "Error in converting subnet prefix to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    
    default:
      for(k = 0; k < i; k++){
	pre = (PREFIX_LIST *)malloc(sizeof(PREFIX_LIST));
	if(!pre)
	  errx("Insufficient memory\n");
	bzero((char *)pre, sizeof(PREFIX_LIST));
	pre->prefix_len = get_and_delete_prefix_len(line[k]);
	if(inet_pton(AF_INET6, line[k], &pre->prefix) <= 0)
	  errx(1, "Error in converting subnet prefix to binary\n");

	*p = pre;
	p = &(*p)->next;
      }
      break;
    }
  }
}

void read_fixed_address(FILE *fp, S_IN6_ADDR_LIST **f, HOST *host) {
  int    i;
  S_IN6_ADDR_LIST *f_addr;

  while(*f)
    f = &(*f)->next;

  printf("Enter twice: ");
  //getchar();
  f_addr = (S_IN6_ADDR_LIST *) malloc(sizeof(S_IN6_ADDR_LIST));
  //getchar();
  if(!f_addr)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) f_addr, sizeof(*f_addr));
  f_addr->prefix_len = get_and_delete_prefix_len(line[1]);
  if(inet_pton(AF_INET6, line[1], &f_addr->ip_addr) <= 0)
    errx(1, "Error in converting host's fixed address to binary\n");

  // Set the values to upper level default 
  f_addr->parent                     = host;
  f_addr->default_preferred_lifetime = ((HOST *)host)->default_preferred_lifetime;
  f_addr->default_valid_lifetime     = ((HOST *)host)->default_valid_lifetime;
  f_addr->max_preferred_lifetime     = ((HOST *)host)->max_preferred_lifetime;
  f_addr->max_valid_lifetime         = ((HOST *)host)->max_valid_lifetime;
  f_addr->scope                      = ((HOST *)host)->scope;
  f_addr->dns_update_A_allowed       = ((HOST *)host)->dns_update_A_allowed;
  f_addr->dns_update_P_allowed       = ((HOST *)host)->dns_update_P_allowed;
  f_addr->domain_name_suffix[0]      = '\0';


  while(1){
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      errx(1, "Unexpected end of file\n");
      break;
      
    case 1:
      if(strcmp(line[0], "domain-name-server{") == 0){
	read_domain_name_server(fp, &f_addr->dn_server_list_head);
      }
      else if(strcmp(line[0], "}") == 0){
	void *parent;

	for(parent = f_addr->parent; parent && !f_addr->dn_server_list_head;
	    parent = ((SUBNET *)parent)->parent) {
	  f_addr->dn_server_list_head = ((SUBNET *)parent)->dn_server_list_head;
	}
	
	for(parent = f_addr->parent; parent && (f_addr->domain_name_suffix[0] == '\0');
	    parent = ((SUBNET *)parent)->parent) {
	  strcpy(f_addr->domain_name_suffix, ((SUBNET *)parent)->domain_name_suffix);
	}
	
	if(f_addr->scope == 0) {
	  /* Set the scope of fixed-addr */
	  if(IN6_IS_ADDR_MULTICAST(&f_addr->ip_addr) || IN6_IS_ADDR_LOOPBACK(&f_addr->ip_addr))
	    warnx("Configuration file has a multicast or loopback address as fixed address"
		  " for fixed-address declaration ending on line %u\n", line_no);
	  if(IN6_IS_ADDR_SITELOCAL(&f_addr->ip_addr))
	    f_addr->scope = 2;
	  if(IN6_IS_ADDR_GLOBAL(&f_addr->ip_addr))
	    f_addr->scope = 1;
	}

	*f = f_addr;
	return;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	f_addr->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	f_addr->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        f_addr->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	f_addr->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	f_addr->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	f_addr->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	f_addr->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "domain-name-server") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_domain_name_server(fp, &f_addr->dn_server_list_head);
      }
      else if(strcmp(line[0], "domain-name-suffix") == 0){
	strcpy(f_addr->domain_name_suffix, line[1]);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
    }
  }
}

RANGE *read_r(FILE *fp, /*RANGE **r,*/ void *i_sub_host) {
  int    i;
  RANGE *range;
  
  //while(*r)
  //  r = &(*r)->next;
  
  range = (RANGE *) malloc(sizeof(RANGE));
  if(!range)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) range, sizeof(RANGE));
  if(inet_pton(AF_INET6, line[1], &range->start) <= 0)
    errx(1, "Error in converting range start address to binary\n");
  
  if(inet_pton(AF_INET6, line[2], &range->end) <= 0)
    errx(1, "Error in converting range end address to binary\n");
  
  /* Set the values to upper level default */
  range->parent                     = i_sub_host;
  range->default_preferred_lifetime = ((SUBNET *)i_sub_host)->default_preferred_lifetime;
  range->default_valid_lifetime     = ((SUBNET *)i_sub_host)->default_valid_lifetime;
  range->max_preferred_lifetime     = ((SUBNET *)i_sub_host)->max_preferred_lifetime;
  range->max_valid_lifetime         = ((SUBNET *)i_sub_host)->max_valid_lifetime;
  range->scope                      = ((SUBNET *)i_sub_host)->scope;
  range->dns_update_A_allowed       = ((SUBNET *)i_sub_host)->dns_update_A_allowed;
  range->dns_update_P_allowed       = ((SUBNET *)i_sub_host)->dns_update_P_allowed;
  range->domain_name_suffix[0]      = '\0';
  range->prefix_len                 = 0;
  //range->preference                 = ((SUBNET *)i_sub_host)->preference;
  //range->server_addr                = ((SUBNET *)i_sub_host)->server_addr;

  while(1){
    i = read_useful_line(fp);
    //if(i != -1)
    //  print_line(i);
    switch(i){
      
    case -1:
      errx(1, "Unexpected end of file\n");
      break;
      
    case 1:
      if(strcmp(line[0], "domain-name-server{") == 0){
	read_domain_name_server(fp, &range->dn_server_list_head);
      }
      else if(strcmp(line[0], "}") == 0){

	void *parent;

	for(parent = range->parent; parent && !range->dn_server_list_head;
	    parent = ((SUBNET *)parent)->parent) {
	  range->dn_server_list_head = ((SUBNET *)parent)->dn_server_list_head;
	}
	
	for(parent = range->parent; parent && (range->domain_name_suffix[0] == '\0');
	      parent = ((SUBNET *)parent)->parent) {
	  strcpy(range->domain_name_suffix, ((SUBNET *)parent)->domain_name_suffix);
	}

	if(range->prefix_len == 0)
	  warnx("prefix-len unspecified or zero for range declaration ending on line %u\n",
		line_no);
	
	/*
	if(IN6_IS_ADDR_MULTICAST(&range->start) || IN6_IS_ADDR_LOOPBACK(&range->start))
	  printf("Configuration file has a multicast or loopback address as start address"
		 " for range declaration ending on line %u\n", line_no);
	
	if(IN6_IS_ADDR_MULTICAST(&range->end) || IN6_IS_ADDR_LOOPBACK(&range->end))
	  printf("Configuration file has a multicast or loopback address as end address"
		 " for range declaration ending on line %u\n", line_no);
	if(!range)
	  printf("...................................NULL\n\n");
	
	if(range->scope == 0) {
	  
	  // We must set the scope to that of start address of range 
	  if(IN6_IS_ADDR_SITELOCAL(&range->start))
	    range->scope = 2;
	  if(IN6_IS_ADDR_GLOBAL(&range->start))
	    range->scope = 1;
	  
	  if(range->scope == 2 && !IN6_IS_ADDR_SITELOCAL(&range->end))
	    warnx("start address and end address have different scopes"
		  " for range declaration ending on line %u\n", line_no);
	  
	  if(range->scope == 1 && !IN6_IS_ADDR_GLOBAL(&range->end))
	    warnx("start address and end address have different scopes"
		  " for range declaration ending on line %u\n", line_no);
	  
	}

	*r = range;
	*/
	return range;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	range->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	range->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        range->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	range->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	range->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	range->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	range->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "domain-name-server") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_domain_name_server(fp, &range->dn_server_list_head);
      }
      else if(strcmp(line[0], "domain-name-suffix") == 0){
	strcpy(range->domain_name_suffix, line[1]);
      }
      else if(strcmp(line[0], "prefix-len") == 0){
	range->prefix_len = atoi(line[1]);
      }
      /*
      else if(strcmp(line[0], "preference") == 0){
        range->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "list-of-prefixes-to-advertise") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_prefixes(fp, &range->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "server-address") == 0){
	(range->server_addr).prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(range->server_addr).prefix) <= 0)
	  errx(1, "Error in converting server address to binary\n");
      }
      */
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
    }
  }
}

void read_range(FILE *fp, RANGE **r, void *i_sub_host) {
  RANGE *range;

  while(*r)
    r = &(*r)->next;

  *r = read_r(fp, i_sub_host);
  range = *r;
  
  if(IN6_IS_ADDR_MULTICAST(&range->start) || IN6_IS_ADDR_LOOPBACK(&range->start))
    printf("Configuration file has a multicast or loopback address as start address"
	   " for range declaration ending on line %u\n", line_no);
  
  if(IN6_IS_ADDR_MULTICAST(&range->end) || IN6_IS_ADDR_LOOPBACK(&range->end))
    printf("Configuration file has a multicast or loopback address as end address"
	   " for range declaration ending on line %u\n", line_no);

  if(range->scope == 0) {
  
    // We must set the scope to that of start address of range 
    if(IN6_IS_ADDR_SITELOCAL(&range->start))
      range->scope = 2;
    if(IN6_IS_ADDR_GLOBAL(&range->start))
      range->scope = 1;
    
    if(range->scope == 2 && !IN6_IS_ADDR_SITELOCAL(&range->end))
      warnx("start address and end address have different scopes"
	    " for range declaration ending on line %u\n", line_no);
    
    if(range->scope == 1 && !IN6_IS_ADDR_GLOBAL(&range->end))
      warnx("start address and end address have different scopes"
	    " for range declaration ending on line %u\n", line_no);
  
    }
  
  
}

void read_host(FILE *fp, HOST **h, void *i_sub){
  HOST           *host;
  int             i, k;
  //S_IN6_ADDR_LIST **t;

  host = (HOST *)malloc(sizeof(HOST));
  if(!host)
    errx(1, "Insufficient memory\n");
  bzero((char *)host, sizeof(HOST));
  
  if(inet_pton(AF_INET6, line[1], &host->link_local_addr) <= 0)
    errx(1, "Error in converting host's link-local-address to binary\n");

  //host->subnet_prefix_len = get_and_delete_prefix_len(line[3]);
  //if(inet_pton(AF_INET6, line[3], &host->subnet_prefix) <= 0)
  //  errx(1, "Error in converting host's subnet-prefix to binary\n");

  /* Set the values to upper level default */
  host->parent                     = i_sub;
  host->default_preferred_lifetime = ((SUBNET *)i_sub)->default_preferred_lifetime;
  host->default_valid_lifetime     = ((SUBNET *)i_sub)->default_valid_lifetime;
  host->max_preferred_lifetime     = ((SUBNET *)i_sub)->max_preferred_lifetime;
  host->max_valid_lifetime         = ((SUBNET *)i_sub)->max_valid_lifetime;
  host->scope                      = ((SUBNET *)i_sub)->scope;
  host->dns_update_A_allowed       = ((SUBNET *)i_sub)->dns_update_A_allowed;
  host->dns_update_P_allowed       = ((SUBNET *)i_sub)->dns_update_P_allowed;
  host->domain_name_suffix[0]      = '\0';
  host->preference                 = ((SUBNET *)i_sub)->preference;
  host->server_addr                = ((SUBNET *)i_sub)->server_addr;
  
  //t = &host->fixed_addr_list_head;

  while(1){
    i = read_useful_line(fp);
    //if(i != -1)
    //  print_line(i);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "list-of-prefixes-to-advertise{") == 0){
	read_prefixes(fp, &host->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "domain-name-server{") == 0){
	read_domain_name_server(fp, &host->dn_server_list_head);
      }
      else if(strcmp(line[0], "}") == 0){

	void *parent;

	for(parent = host->parent; parent && !host->dn_server_list_head;
	    parent = ((SUBNET *)parent)->parent) {
	  host->dn_server_list_head = ((SUBNET *)parent)->dn_server_list_head;
	}
	
	for(parent = host->parent; parent && (host->domain_name_suffix[0] == '\0');
	      parent = ((SUBNET *)parent)->parent) {
	  strcpy(host->domain_name_suffix, ((SUBNET *)parent)->domain_name_suffix);
	}

	if(!host->adv_prefix_list_head)
	  host->adv_prefix_list_head = ((SUBNET *)host->parent)->adv_prefix_list_head;

	if(!host->range_list_head)
	  host->range_list_head = ((SUBNET *)host->parent)->range_list_head;

	while(*h)
	  h = &(*h)->next;
	*h = host;
	return;
      }
      break;
    
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	host->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	host->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        host->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	host->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	host->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	host->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	host->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "domain-name-server") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_domain_name_server(fp, &host->dn_server_list_head);
      }
      else if(strcmp(line[0], "domain-name-suffix") == 0){
	strcpy(host->domain_name_suffix, line[1]);
      }
      else if(strcmp(line[0], "preference") == 0){
	host->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "list-of-prefixes-to-advertise") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_prefixes(fp, &host->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "server-address") == 0){
	(host->server_addr).prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(host->server_addr).prefix) <= 0)
	  errx(1, "Error in converting server address to binary\n");
      }
      else if(strcmp(line[0], "fixed-address") == 0 &&
	      line[1][strlen(line[1]) - 1] == '{'){
	line[1][strlen(line[1]) - 1] = '\0';
	read_fixed_address(fp, &host->fixed_addr_list_head, host);
	/*
	 *t = (S_IN6_ADDR_LIST *)malloc(sizeof(S_IN6_ADDR_LIST));
	 if(!(*t))
	 errx(1, "Insufficient memory\n");
	 bzero((char *)(*t), sizeof(S_IN6_ADDR_LIST));
	 //(*t)->scope      = get_and_delete_scope(line[1]);
	 (*t)->prefix_len = get_and_delete_prefix_len(line[1]);
	 if(inet_pton(AF_INET6, line[1], &(*t)->ip_addr) <= 0)
	 errx(1, "Error in converting host's fixed address to binary\n");
	 
	 t = &(*t)->next;
	*/
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 3:
      if(strcmp(line[0], "range") == 0 && line[2][strlen(line[2]) - 1] == '{'){
	line[2][strlen(line[2]) - 1] = '\0';
	/* Note: First goto end of list of range and then add */
	read_range(fp, &host->range_list_head, host);
      }
      else if(strcmp(line[0], "fixed-address") == 0 &&
	      strcmp(line[2], "{") == 0){
	read_fixed_address(fp, &host->fixed_addr_list_head, host);
	/*
	  for(k = 1; k < i; k++){
	  *t = (S_IN6_ADDR_LIST *)malloc(sizeof(S_IN6_ADDR_LIST));
	  if(!(*t))
	  errx(1, "Insufficient memory\n");
	  bzero((char *)(*t), sizeof(S_IN6_ADDR_LIST));
	  (*t)->scope      = get_and_delete_scope(line[k]);
	  (*t)->prefix_len = get_and_delete_prefix_len(line[k]);
	  if(inet_pton(AF_INET6, line[k], &(*t)->ip_addr) <= 0)
	  errx(1, "Error in converting host's fixed address to binary\n");
	  
	  t = &(*t)->next;
	*/
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;

    case 4:
      if(strcmp(line[0], "range") == 0 && strcmp(line[3], "{") == 0){
	/* Note: First goto end of list of range and then add */
	read_range(fp, &host->range_list_head, host);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;

    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
    }
  }
}

void read_interface(FILE *fp, INTERFACE *interface){
  int i;
  
  while(1){
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      errx(1, "Unexpected end of file\n");
      break;
      
    case 1:
      if(strcmp(line[0], "list-of-prefixes-to-advertise{") == 0){
	read_prefixes(fp, &interface->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "domain-name-server{") == 0){
	read_domain_name_server(fp, &interface->dn_server_list_head);
      }
      else if(strcmp(line[0], "}") == 0){
	return;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	interface->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	interface->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        interface->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	interface->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	interface->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	interface->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	interface->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "domain-name-server") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_domain_name_server(fp, &interface->dn_server_list_head);
      }
      else if(strcmp(line[0], "domain-name-suffix") == 0){
	strcpy(interface->domain_name_suffix, line[1]);
      }
      else if(strcmp(line[0], "preference") == 0){
	interface->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "list-of-prefixes-to-advertise") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_prefixes(fp, &interface->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "server-address") == 0){
	(interface->server_addr).prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(interface->server_addr).prefix) <= 0)
	  errx(1, "Error in converting server address to binary\n");
      }
      else if(strcmp(line[0], "host") == 0 &&
	      line[1][strlen(line[1]) - 1] == '{'){
	line[1][strlen(line[1]) - 1] = '\0';
	read_host(fp, &interface->host_list_head, interface);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 3:
      if(strcmp(line[0], "range") == 0 && line[2][strlen(line[2]) - 1] == '{'){
	line[2][strlen(line[2]) - 1] = '\0';
	/* Note: First goto end of list of range and then add */
	read_range(fp, &interface->range_list_head, interface);
      }
      else if(strcmp(line[0], "host") == 0 &&
	 strcmp(line[2], "{") == 0){
	read_host(fp, &interface->host_list_head, interface);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 4:
      if(strcmp(line[0], "range") == 0 && strcmp(line[3], "{") == 0){
	/* Note: First goto end of list of range and then add */
	read_range(fp, &interface->range_list_head, interface);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;

    case 5:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    }
  }
}

SUBNET *read_subnet(FILE *fp, int i){
  int         k;
  SUBNET     *subnet;
  RANGE     **r;
  IN6_ADDR_WITH_PREFIX **t;
   
  subnet = (SUBNET *) malloc(sizeof(SUBNET));
  if(!subnet)
    errx(1, "Insufficient memory\n");
  
  bzero((char *) subnet, sizeof(SUBNET));
  /* Default values for dns update of A and P records */
  subnet->dns_update_A_allowed = subnet->dns_update_A_allowed = 1;
  t = &subnet->addr_list_head;

  for(k = 1; k <= i; k++){
    *t = (IN6_ADDR_WITH_PREFIX *)malloc(sizeof(IN6_ADDR_WITH_PREFIX));
    if(!(*t))
      errx(1, "Insufficient memory\n");
    bzero((char *)(*t), sizeof(IN6_ADDR_WITH_PREFIX));
    (*t)->prefix_len = get_and_delete_prefix_len(line[k]);
    if(inet_pton(AF_INET6, line[k], &(*t)->prefix) <= 0)
      errx(1, "Error in converting subnet prefix to binary\n");
    t = &(*t)->next;
  }

  while(1){
    i = read_useful_line(fp);
    switch(i){

    case -1:
      errx(1, "Unexpected end of file\n");
      break;

    case 1:
      if(strcmp(line[0], "list-of-prefixes-to-advertise{") == 0){
	read_prefixes(fp, &subnet->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "domain-name-server{") == 0){
	read_domain_name_server(fp, &subnet->dn_server_list_head);
      }
      else if(strcmp(line[0], "}") == 0){
	return subnet;
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(strcmp(line[0], "default-preferred-lifetime") == 0){
	subnet->default_preferred_lifetime = atoi(line[1]);	
      }
      else if(strcmp(line[0], "default-valid-lifetime") == 0){
	subnet->default_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-preferred-lifetime") == 0){
        subnet->max_preferred_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "max-valid-lifetime") == 0){
	subnet->max_valid_lifetime = atoi(line[1]);
      }
      else if(strcmp(line[0], "scope") == 0){
	subnet->scope = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-A-allowed") == 0){
	subnet->dns_update_A_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "dns-update-P-allowed") == 0){
	subnet->dns_update_P_allowed = atoi(line[1]);
      }
      else if(strcmp(line[0], "domain-name-server") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_domain_name_server(fp, &subnet->dn_server_list_head);
      }
      else if(strcmp(line[0], "domain-name-suffix") == 0){
	strcpy(subnet->domain_name_suffix, line[1]);
      }
      else if(strcmp(line[0], "preference") == 0){
	subnet->preference = atoi(line[1]);
      }
      else if(strcmp(line[0], "list-of-prefixes-to-advertise") == 0 &&
	      strcmp(line[1], "{") == 0){
	read_prefixes(fp, &subnet->adv_prefix_list_head);
      }
      else if(strcmp(line[0], "server-address") == 0){
	(subnet->server_addr).prefix_len = get_and_delete_prefix_len(line[1]);
	if(inet_pton(AF_INET6, line[1], &(subnet->server_addr).prefix) <= 0)
	  errx(1, "Error in converting server address to binary\n");
      }
      else if(strcmp(line[0], "host") == 0 &&
	      line[1][strlen(line[1]) - 1] == '{'){
	line[1][strlen(line[1]) - 1] = '\0';
	read_host(fp, &subnet->host_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 3:
      if(strcmp(line[0], "range") == 0 && line[2][strlen(line[2]) - 1] == '{'){
	line[2][strlen(line[2]) - 1] = '\0';
	/* Note: First goto end of list of range and then add */
	read_range(fp, &subnet->range_list_head, subnet);
      }
      else if(strcmp(line[0], "host") == 0 &&
	 strcmp(line[2], "{") == 0){
	read_host(fp, &subnet->host_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 4:
      if(strcmp(line[0], "range") == 0 && strcmp(line[3], "{") == 0){
	/* Note: First goto end of list of range and then add */
	read_range(fp, &subnet->range_list_head, subnet);
      }
      else
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      break;

    case 5:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    default:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    }
  }
  warnx("This should be impossible\n");
  return subnet;
}

void print_dn_server_list(IN6_ADDR_WITH_PREFIX *head) {

  printf("dn-server-list: ");
  for(; head; head = head->next){
    printf("%s/%u ", inet_ntop(AF_INET6, &head->prefix, name, 64), head->prefix_len);
  }
  printf("\n\t");
}

void print_fixed_address_list(S_IN6_ADDR_LIST *head) {

  for(; head; head = head->next) {
    printf("\nfixed address ");
    printf("%s/%u\n\t", inet_ntop(AF_INET6, &head->ip_addr, name, 64),
	   head->prefix_len);
    printf("default-preferred-lifetime: %u\n\t", head->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n\t", head->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n\t", head->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n\t", head->max_valid_lifetime);
    printf("scope: %u\n\t", head->scope);
    printf("dns-update-A-allowed: %u\n\t", head->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\t", head->dns_update_P_allowed);
    print_dn_server_list(head->dn_server_list_head);
    printf("domain-name-suffix: %s\n\t", head->domain_name_suffix);
      
  }

}

void print_interface(INTERFACE *interface){
  IN6_ADDR_WITH_PREFIX *temp_addr_pre;
  S_IN6_ADDR_LIST      *temp_addr;
  IN6_ADDR_LIST        *temp_iaddr;
  PREFIX_LIST          *temp_pre;
  RANGE                *temp_range;
  HOST                 *temp_host;

  printf("\n****************************************************\n");
  printf("interface %s index = %u\n\t", interface->name, interface->index);
  printf("link-local-addr = %s\n\t", inet_ntop(AF_INET6, interface->link_local_addr, name, 64));
  printf("IPv6-addr-list: ");
  for(temp_iaddr = interface->addr_list_head; temp_iaddr; temp_iaddr = temp_iaddr->next) {
    printf("%s ", inet_ntop(AF_INET6, &temp_iaddr->ip_addr, name, 64));
  }
  printf("\n\t");
  printf("default-preferred-lifetime: %u\n\t", interface->default_preferred_lifetime);
  printf("default-valid-lifetime: %u\n\t", interface->default_valid_lifetime);
  printf("max-preferred-lifetime: %u\n\t", interface->max_preferred_lifetime);
  printf("max-valid-lifetime: %u\n\t", interface->max_valid_lifetime);
  printf("scope: %u\n\t", interface->scope);
  printf("dns-update-A-allowed: %u\n\t", interface->dns_update_A_allowed);
  printf("dns-update-P-allowed: %u\n\t", interface->dns_update_P_allowed);
  print_dn_server_list(interface->dn_server_list_head);
  printf("domain-name-suffix: %s\n\t", interface->domain_name_suffix);
  printf("preference: %u\n\n\t", interface->preference);

  printf("list-of-prefixes-to-advertise: ");
  for(temp_pre = interface->adv_prefix_list_head; temp_pre; temp_pre = temp_pre->next){
    printf("%s/%u ", inet_ntop(AF_INET6, &temp_pre->prefix, name, 64), temp_pre->prefix_len);
  }
  printf("\n\n\t");
  printf("server-address: %s/%u\n",
	 inet_ntop(AF_INET6, &(interface->server_addr).prefix, name, 64),
	 (interface->server_addr).prefix_len);

  /* Print host declarations */
  for(temp_host = interface->host_list_head; temp_host; temp_host = temp_host->next){
    printf("\nhost %s\n\t", inet_ntop(AF_INET6, &temp_host->link_local_addr, name, 64));

    printf("default-preferred-lifetime: %u\n\t", temp_host->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n\t", temp_host->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n\t", temp_host->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n\t", temp_host->max_valid_lifetime);
    printf("scope: %u\n\t", temp_host->scope);
    printf("dns-update-A-allowed: %u\n\t", temp_host->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\t", temp_host->dns_update_P_allowed);
    print_dn_server_list(temp_host->dn_server_list_head);
    printf("domain-name-suffix: %s\n\t", temp_host->domain_name_suffix);
    printf("preference: %u\n\n\t", temp_host->preference);

    printf("list-of-prefixes-to-advertise: ");
    for(temp_pre = temp_host->adv_prefix_list_head; temp_pre; temp_pre = temp_pre->next){
      printf("%s/%u ", inet_ntop(AF_INET6, &temp_pre->prefix, name, 64), temp_pre->prefix_len);
    }
    printf("\n\n\t");
    printf("server-address: %s/%u\n",
	   inet_ntop(AF_INET6, &(temp_host->server_addr).prefix, name, 64),
	   (temp_host->server_addr).prefix_len);
    print_fixed_address_list(temp_host->fixed_addr_list_head);
    printf("\n\n\t");
    /* Print range declarations */
    for(temp_range = temp_host->range_list_head; temp_range; temp_range = temp_range->next){
      printf("\nrange %s ",
	     inet_ntop(AF_INET6, &temp_range->start, name, 64));
      printf("%s\n\t", inet_ntop(AF_INET6, &temp_range->end, name, 64));
      printf("default-preferred-lifetime: %u\n\t", temp_range->default_preferred_lifetime);
      printf("default-valid-lifetime: %u\n\t", temp_range->default_valid_lifetime);
      printf("max-preferred-lifetime: %u\n\t", temp_range->max_preferred_lifetime);
      printf("max-valid-lifetime: %u\n\t", temp_range->max_valid_lifetime);
      printf("scope: %u\n\t", temp_range->scope);
      printf("dns-update-A-allowed: %u\n\t", temp_range->dns_update_A_allowed);
      printf("dns-update-P-allowed: %u\n\n\t", temp_range->dns_update_P_allowed);
      print_dn_server_list(temp_range->dn_server_list_head);
      printf("domain-name-suffix: %s\n\t", temp_range->domain_name_suffix);
      printf("prefix-len: %u\n\n\t", temp_range->prefix_len);
    }
  }
  /* Print range declarations */
  for(temp_range = interface->range_list_head; temp_range; temp_range = temp_range->next){
    printf("\nrange %s ",
	   inet_ntop(AF_INET6, &temp_range->start, name, 64));
    printf("%s\n\t", inet_ntop(AF_INET6, &temp_range->end, name, 64));
    printf("default-preferred-lifetime: %u\n\t", temp_range->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n\t", temp_range->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n\t", temp_range->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n\t", temp_range->max_valid_lifetime);
    printf("scope: %u\n\t", temp_range->scope);
    printf("dns-update-A-allowed: %u\n\t", temp_range->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\n\t", temp_range->dns_update_P_allowed);
    print_dn_server_list(temp_range->dn_server_list_head);
    printf("domain-name-suffix: %s\n\t", temp_range->domain_name_suffix);
    printf("prefix-len: %u\n\n", temp_range->prefix_len);
  }
}

void print_subnet(SUBNET *subnet){
  IN6_ADDR_WITH_PREFIX *temp_addr_pre;
  S_IN6_ADDR_LIST      *temp_addr;
  PREFIX_LIST          *temp_pre;
  RANGE                *temp_range;
  HOST                 *temp_host;

  printf("\n****************************************************\n");
  printf("subnet: ");
  for(temp_addr_pre = subnet->addr_list_head; temp_addr_pre;
      temp_addr_pre = temp_addr_pre->next){
    printf("%s/%u ",inet_ntop(AF_INET6, &temp_addr_pre->prefix, name, 64),
	   temp_addr_pre->prefix_len);
  }
  printf("\n\t");
  /* Print top level variables */
  printf("default-preferred-lifetime: %u\n\t", subnet->default_preferred_lifetime);
  printf("default-valid-lifetime: %u\n\t", subnet->default_valid_lifetime);
  printf("max-preferred-lifetime: %u\n\t", subnet->max_preferred_lifetime);
  printf("max-valid-lifetime: %u\n\t", subnet->max_valid_lifetime);
  printf("scope: %u\n\t", subnet->scope);
  printf("dns-update-A-allowed: %u\n\t", subnet->dns_update_A_allowed);
  printf("dns-update-P-allowed: %u\n\t", subnet->dns_update_P_allowed);
  print_dn_server_list(subnet->dn_server_list_head);
  printf("domain-name-suffix: %s\n\t", subnet->domain_name_suffix);
  printf("preference: %u\n\n\t", subnet->preference);

  printf("list-of-prefixes-to-advertise: ");
  for(temp_pre = subnet->adv_prefix_list_head; temp_pre; temp_pre = temp_pre->next){
    printf("%s/%u ", inet_ntop(AF_INET6, &temp_pre->prefix, name, 64), temp_pre->prefix_len);
  }
  printf("\n\n\t");
  printf("server-address: %s/%u\n\t",
	 inet_ntop(AF_INET6, &(subnet->server_addr).prefix, name, 64),
	 (subnet->server_addr).prefix_len);

  /* Print host declarations */
  for(temp_host = subnet->host_list_head; temp_host; temp_host = temp_host->next){
    printf("\nhost %s\n\t", inet_ntop(AF_INET6, &temp_host->link_local_addr, name, 64));
    //printf("subnet-prefix %s/%u\n",
    //   inet_ntop(AF_INET6, &temp_host->subnet_prefix, name, 64),
    //   temp_host->subnet_prefix_len);

    printf("default-preferred-lifetime: %u\n\t", temp_host->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n\t", temp_host->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n\t", temp_host->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n\t", temp_host->max_valid_lifetime);
    printf("scope: %u\n\t", temp_host->scope);
    printf("dns-update-A-allowed: %u\n\t", temp_host->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\t", temp_host->dns_update_P_allowed);
    print_dn_server_list(temp_host->dn_server_list_head);
    printf("domain-name-suffix: %s\n\t", temp_host->domain_name_suffix);
    printf("preference: %u\n\n\t", temp_host->preference);

    printf("list-of-prefixes-to-advertise: ");
    for(temp_pre = temp_host->adv_prefix_list_head; temp_pre; temp_pre = temp_pre->next){
      printf("%s/%u ", inet_ntop(AF_INET6, &temp_pre->prefix, name, 64), temp_pre->prefix_len);
    }
    printf("\n\n\t");
    printf("server-address: %s/%u\n\t",
	   inet_ntop(AF_INET6, &(temp_host->server_addr).prefix, name, 64),
	   (temp_host->server_addr).prefix_len);
    print_fixed_address_list(temp_host->fixed_addr_list_head);
    printf("\n\n");
    /* Print range declarations */
    for(temp_range = temp_host->range_list_head; temp_range; temp_range = temp_range->next){
      printf("\nrange %s ",
	     inet_ntop(AF_INET6, &temp_range->start, name, 64));
      printf("%s\n\t", inet_ntop(AF_INET6, &temp_range->end, name, 64));
      printf("default-preferred-lifetime: %u\n\t", temp_range->default_preferred_lifetime);
      printf("default-valid-lifetime: %u\n\t", temp_range->default_valid_lifetime);
      printf("max-preferred-lifetime: %u\n\t", temp_range->max_preferred_lifetime);
      printf("max-valid-lifetime: %u\n\t", temp_range->max_valid_lifetime);
      printf("scope: %u\n\t", temp_range->scope);
      printf("dns-update-A-allowed: %u\n\t", temp_range->dns_update_A_allowed);
      printf("dns-update-P-allowed: %u\n\n\t", temp_range->dns_update_P_allowed);
      print_dn_server_list(temp_range->dn_server_list_head);
      printf("domain-name-suffix: %s\n\t", temp_range->domain_name_suffix);
      printf("prefix-len: %u\n\n\t", temp_range->prefix_len);
    }
  }
  /* Print range declarations */
  for(temp_range = subnet->range_list_head; temp_range; temp_range = temp_range->next){
    printf("\nrange %s ",
	   inet_ntop(AF_INET6, &temp_range->start, name, 64));
    printf("%s\n\t", inet_ntop(AF_INET6, &temp_range->end, name, 64));
    printf("default-preferred-lifetime: %u\n\t", temp_range->default_preferred_lifetime);
    printf("default-valid-lifetime: %u\n\t", temp_range->default_valid_lifetime);
    printf("max-preferred-lifetime: %u\n\t", temp_range->max_preferred_lifetime);
    printf("max-valid-lifetime: %u\n\t", temp_range->max_valid_lifetime);
    printf("scope: %u\n\t", temp_range->scope);
    printf("dns-update-A-allowed: %u\n\t", temp_range->dns_update_A_allowed);
    printf("dns-update-P-allowed: %u\n\t", temp_range->dns_update_P_allowed);
    print_dn_server_list(temp_range->dn_server_list_head);
    printf("domain-name-suffix: %s\n\t", temp_range->domain_name_suffix);
    printf("prefix-len: %u\n\n\t", temp_range->prefix_len);
  }
}

INTERFACE *find_interface_by_name(char *name, INTERFACE *i_head){
  while(i_head){
    if(strcmp(i_head->name, name) == 0)
      return i_head;
    i_head = i_head->next;
  }
  return NULL;
}

void parse_address_properties_file(char *file, SUBNET **s_head, INTERFACE *i_head){
  FILE       *fp;
  int         i, k;
  SUBNET    **subnet = s_head;
  INTERFACE  *temp_interface;
  
  line_no = 0;/* 0 lines read */
  ADDRESS_PROPERTIES_FILE = file;
  fp = fopen(ADDRESS_PROPERTIES_FILE, "r");
  if(!fp)
    errx(1, "Error opening %s. Exiting...\n", ADDRESS_PROPERTIES_FILE);
  
  while(1){ 
    i = read_useful_line(fp);
    switch(i){
      
    case -1:
      /* EOF reached */
      return;
      break;
      
    case 1:
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
      break;
      
    case 2:
      if(line[1][strlen(line[1]) - 1] != '{')
	errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	     ADDRESS_PROPERTIES_FILE, line_no);
      
      line[1][strlen(line[1]) - 1] = '\0';
      break;
      
    default:
      break;
    }

    i--;
    if(strcmp(line[i], "{") == 0)
      i--;
    else if(line[i][strlen(line[i]) - 1] == '{')
      line[i][strlen(line[i]) - 1] = '\0';
    else
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
    
    if(strcmp(line[0], "subnet") == 0){
      *subnet = read_subnet(fp, i);
      if(!(*subnet))
	break;
      subnet = &(*subnet)->next;
    }
    else if(strcmp(line[0], "interface") == 0){
      temp_interface = find_interface_by_name(line[1], i_head);
      if(!temp_interface)
	errx(1, "Erroneous configuration file %s. Couldn't find interface %s."
	     " Successfully read %u lines\n", ADDRESS_PROPERTIES_FILE, line[1], line_no);
      read_interface(fp, temp_interface);
      //print_interface(interface);
    }
    else
      errx(1, "Erroneous configuration file %s. Successfully read %u lines\n",
	   ADDRESS_PROPERTIES_FILE, line_no);
    
  }
}
/*
main(int argc, char *argv[]){
  INTERFACE *in_list_head = NULL;
  SUBNET *subnet_list_head;
  INTERFACE  **temp, *t;
  SUBNET *sub;

  if(argc != 2){
    printf("Incorrect number of arguments to main\n");
    exit(0);
  }
  discover_interfaces(&in_list_head);
  getchar();
  print(&in_list_head);
  getchar();
  
  parse_address_properties_file(argv[1], &subnet_list_head, in_list_head);
  for(sub = subnet_list_head; sub; sub = sub->next){
    print_subnet(sub);
  }
  for(t = in_list_head; t; t = t->next){
    print_interface(t);
  }
 
}
*/
