/* Implemented by Dhawal Kumar */
# include <stdio.h>
# include <errno.h>
# include <signal.h>
# include <netdb.h>
# include <string.h>
# include <stdlib.h>
# include <syslog.h>

# include <sys/types.h>
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/param.h>
# include <sys/file.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/uio.h>

# include <netinet/in.h>

# include <arpa/inet.h>

# include <net/if.h>
# include <net/if_dl.h>
# include <net/route.h>
// Remove the following include. It is used only for sleep
#include <unistd.h>
//
# include "defines.h"
typedef struct timeval TIME;

# define PACKET
# define CURRENT_TIME
# define TIMEOUT
/* Utilities */
# include "../common/time_utility.h"
# include "../common/get_set_bits_utility.h"
# include "../common/read_write_data.h"
# include "../common/print_packet.h"
# include "in6_utility.h"
# include "log_error.h"

/* Define the structures here */
# include "../common/dhcp_messages.h"
# include "../common/ip_settings.h"
# include "in6_addr_list.h"
# include "host_subnet_interface_structures.h"
# include "lease_structure.h"
# include "delete_expired_leases.h"

int has_expired(LEASE *lease) {
  TIME t;

  if(gettimeofday(&t, NULL) == -1)
    errx(1, "Error in gettimeofday. Exiting...\n");
  printf("\t...Checking expiry for %d %d at time %d %d", lease->when_leased.tv_sec, lease->when_leased.tv_usec, t.tv_sec, t.tv_usec);
  printf("\t...diff in micro %d %d\n", DIFF_IN_MICRO(t, lease->when_leased),
	 lease->valid_lifetime * 1000000);
  return (DIFF_IN_MS(t, lease->when_leased) >= (lease->valid_lifetime * 1000));
}


/* Globals */
char *recv_msg, *send_msg;
int reply_fd;
INTERFACE           *in_list_head;
SUBNET              *subnet_list_head;
LEASE               *lease_list_head;

extern INTERFACE *find_interface(uint index, INTERFACE *head);

/* functions used */
# include "dns_operations.h"
//# include "get_new_addr.h"
# include "handle_packet_from_client.h"

#ifdef PRINT
#undef PRINT
#endif

main(int argc, char *argv[]){
  int                  agent_fd, administrator_fd, adm_src_fd = -1, on;
  int                  nfds = -1, n, childpid, adm_src_socklen;
  uint32_t             main_interface, in_multi_if, out_multi_if;
  char                 name[64], adm_buf[ADM_BUF_SIZE];
  struct sockaddr_in6  agent_sockaddr, administrator_sockaddr, adm_src_sockaddr;
  struct sockaddr_in6  cli_sockaddr, reply_sockaddr;
  /*
   * Note: reply sockaddr is used for reply_fd binding while cli_sockaddr is used
   * for replying back to clients. It holds CLIENT_PORT as destination port no.
   */
  fd_set               f, temp_f;
  TIME                 timeout;
  struct ipv6_mreq     socket_opt;
  struct msghdr        msg;
  struct iovec         iov[1];
  struct{
    struct cmsghdr     c_cm;
    struct in6_pktinfo c_pi;
  }c_cmsg;
  
#define pi	      c_cmsg.c_pi
  
  
  /* Note: administrator_sockfd is used for listening to administrator's
   * reconfigure messages on ADMINISTRATOR_PORT. This is a TCP socket.
   */
  /* Imp: See below. Its not yet done. Read from conf file or through cmd line options */
  in_multi_if    = 1;
  out_multi_if   = 1;
  main_interface = 1;
  
  if(argc > 1 && strcmp(argv[1], "-d") == 0) {
    childpid = fork();
    
    if(childpid < 0)
      errx(1, "Error in fork. Exiting...");
    
    if(childpid != 0)
      exit(0);/* Parent exits */
    
    printf("Child continuing\n");
  }

  recv_msg = (char *) malloc(MAX_DHCP_MSG);
  send_msg = (char *) malloc(MAX_DHCP_MSG);

  FD_ZERO(&f);
  FD_ZERO(&temp_f);

  discover_interfaces(&in_list_head);
  //getchar();
  fflush(stdout);
  if(argc > 1)
    parse_address_properties_file(argv[1], &subnet_list_head, in_list_head);
  else
    parse_address_properties_file("parse/properties.txt", &subnet_list_head, in_list_head);
#ifdef PRINT
  {
    SUBNET    *sub;
    INTERFACE *t;
    for(sub = subnet_list_head; sub; sub = sub->next){
      print_subnet(sub);
      //fflush(stdout);
    }
    for(t = in_list_head; t; t = t->next){
      print_interface(t);
      //fflush(stdout);
      //getchar();
    }
  }
#endif

  if(argc > 2)
    read_lease_file(argv[2], &lease_list_head);
  else
    read_lease_file("parse/lease.txt", &lease_list_head);

#ifdef PRINT
  {
    LEASE *lease;
    for(lease = lease_list_head; lease; lease = lease->next)
      print_lease(lease);
    //fflush(stdout);
  }
#endif
  //fflush(stdout);
  //getchar();
  //exit(0);
  /* Open all the required sockets and bind them                         */
  /* agent_fd        : "FF02::1:2", "FF05::1:3", in6addr_any, AGENT_PORT */
  /* administrator_fd: in6addr_any, ADMINISTRATOR_PORT                   */

  agent_fd = socket(AF_INET6, SOCK_DGRAM, 0);
  if(agent_fd < 0)
    errx(1, "Error in opening datagram socket. Exiting...");

  /* socket_opt.ipv6mr_multiaddr = "FF05::1:3"; */
  n = inet_pton(AF_INET6, ALL_DHCP_SERVERS_ADDR,
                & socket_opt.ipv6mr_multiaddr); 
  on = 1;
  if (setsockopt(agent_fd, IPPROTO_IPV6, IPV6_RECVPKTINFO,
		 &on, sizeof(on)) == -1)
    errx(1, "Couldn't set socket option to IPV6_RECVPKTINFO. Exiting...");

  if(n <= 0){
    if(n < 0)
      errx(1, "Unknown AF(first) argument. Exiting...");
    else
      errx(1, "Invalid address string. Exiting...");
  }

  socket_opt.ipv6mr_interface = in_multi_if;
  n = setsockopt(agent_fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, (char *) &socket_opt, sizeof(socket_opt));
  if(n == -1){
    printf("Couldn't set socket option to IPV6_JOIN_GROUP to %s. ",
	   inet_ntop(AF_INET6, &socket_opt.ipv6mr_multiaddr, name, 64));
    errx(1, "Exiting...");
  }
  
  /* socket_opt.ipv6mr_multiaddr = "FF02::1:2"; */
  n = inet_pton(AF_INET6, ALL_DHCP_AGENTS_ADDR,
                & socket_opt.ipv6mr_multiaddr);   
  if(n <= 0){
    if(n < 0)
      errx(1, "Unknown AF(first) argument. Exiting...");
    else
      errx(1, "Invalid address string. Exiting...");
  }
  
  socket_opt.ipv6mr_interface = in_multi_if;
  n = setsockopt(agent_fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, (char *) &socket_opt, sizeof(socket_opt));
  if(n == -1){
    printf("Couldn't set socket option to IPV6_JOIN_GROUP to %s. ",
	   inet_ntop(AF_INET6, &socket_opt.ipv6mr_multiaddr, name, 64));
    errx(1, "Exiting...");
  }
  
  administrator_fd = socket(AF_INET6, SOCK_STREAM, 0);
  if(administrator_fd < 0)
    errx(1, "Error in opening stream socket. Exiting...");
 
  reply_fd = socket(AF_INET6, SOCK_DGRAM, 0);
  if(reply_fd < 0)
    errx(1, "Error in opening datagram socket. Exiting...");
  /* Set sockopt to  IPV6_MULTICAST_IF for DHCP reconfigure messages */
  if(setsockopt(reply_fd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		 &out_multi_if, sizeof(out_multi_if)) < 0){
    errx(1, "Couldn't set socket option to  IPV6_MULTICAST_IF. Exiting...");
  }
  
  INITIALISE_SOCKADDR(agent_sockaddr);
#ifdef SIN6_LEN
  agent_sockaddr.sin6_len  = sizeof(agent_sockaddr);
#endif
  agent_sockaddr.sin6_port = htons(AGENT_PORT);
  agent_sockaddr.sin6_addr = in6addr_any;
  if((n = bind(agent_fd, (struct sockaddr *) &agent_sockaddr, 
	       sizeof(agent_sockaddr))) < 0){
    printf("bind: %d\a\n",n);
    errx(1, "Error while binding address! Exiting...");
  }

  INITIALISE_SOCKADDR(administrator_sockaddr);
#ifdef SIN6_LEN
  administrator_sockaddr.sin6_len  = sizeof(administrator_sockaddr);
#endif
  administrator_sockaddr.sin6_port = htons(ADMINISTRATOR_PORT);
  administrator_sockaddr.sin6_addr = in6addr_any;
  if((n = bind(administrator_fd, (struct sockaddr *) &administrator_sockaddr, 
	       sizeof(administrator_sockaddr))) < 0){
    printf("bind: %d\a\n",n);
    errx(1, "Error while binding address! Exiting...");
  }

  INITIALISE_SOCKADDR(reply_sockaddr);
#ifdef SIN6_LEN
  reply_sockaddr.sin6_len  = sizeof(reply_sockaddr);
#endif
  reply_sockaddr.sin6_port = htons(0);
  reply_sockaddr.sin6_addr = in6addr_any;
  if((n = bind(reply_fd, (struct sockaddr *) &reply_sockaddr, 
	       sizeof(reply_sockaddr))) < 0){
    printf("bind: %d\a\n",n);
    errx(1, "Error while binding address! Exiting...");
  }
  
  FD_SET(agent_fd, &f);
  nfds = agent_fd;

  FD_SET(administrator_fd, &f);
  nfds = (nfds > administrator_fd) ? nfds : administrator_fd;

  temp_f = f;

  /* Initialise data structures for receiving ancillary data */
  iov[0].iov_base = recv_msg;
  iov[0].iov_len  = MAX_DHCP_MSG;
  msg.msg_name    = (caddr_t) &cli_sockaddr;
  msg.msg_namelen = sizeof(cli_sockaddr);
  msg.msg_iov     = &iov[0];
  msg.msg_iovlen  = 1;
  msg.msg_control = (caddr_t)&c_cmsg;
  msg.msg_controllen = sizeof(c_cmsg);

#ifdef CURRENT_TIME  
  if(gettimeofday(&timeout, NULL) == -1)
    errx("Error in gettimeofday. Exiting...\n");

  printf("Current time %u %u\n", timeout.tv_sec, timeout.tv_usec);
#endif
  
  SET_TIMEOUT(timeout, SLEEP_TIME * 100);
  listen(administrator_fd, 5);


#ifdef DEBUG
  fflush(stdout);
  //getchar();
  printf("agent_fd[%u] addr %s port %u nfds %u\n", agent_fd, 
	 inet_ntop(AF_INET6, &agent_sockaddr.sin6_addr, name, 64), ntohs(agent_sockaddr.sin6_port), nfds);
#endif

  while(1){
#ifdef TIMEOUT
    printf("Sleeping for %d %d\n", timeout.tv_sec, timeout.tv_usec);
#endif
    n = select(nfds + 1, &f, (fd_set *)0, (fd_set *)0, &timeout);
#ifdef DEBUG
    printf("select = %d\n", n);
#endif
    
    if(n < 0)
      errx(1, "Error in select\a\n");
    
    if(n == 0){
      timeout = delete_expired_leases(&lease_list_head);
      f = temp_f;
      continue;
    }
    if(FD_ISSET(agent_fd, &f)){
      /* Note: Only solicit messages are multicast???? */
      FD_CLR(agent_fd, &f);
      n = recvmsg(agent_fd, &msg, 0);
#ifdef PACKET
      printf("Received from %s\n", inet_ntop(AF_INET6, &cli_sockaddr.sin6_addr, name, 64));
      print_packet(recv_msg, n);
#endif
      handle_packet_from_client(c_cmsg.c_pi.ipi6_ifindex, recv_msg, n, &cli_sockaddr);//, MULTICAST_PACKET);
    }

    else if(FD_ISSET(administrator_fd, &f)){
      /* This means a connect request from administrator has arrived */
      /* Check if no administrator is already talking and accept     */
      FD_CLR(administrator_fd, &f);

      if(adm_src_fd != -1){
	printf("Two administrators can't simultaneously talk. Accepting new one. Be careful\n");
      }
      adm_src_socklen = sizeof(adm_src_sockaddr);
      adm_src_fd = accept(administrator_fd, (struct sockaddr *) &adm_src_sockaddr, &adm_src_socklen);
      if(adm_src_fd < 0)
	errx(1, "Accept error. Exiting...");

      FD_SET(adm_src_fd, &temp_f);
    }

    else if(adm_src_fd >= 0 && FD_ISSET(adm_src_fd, &f)){
      /* Administrator has sent some message */
      /* Imp: This will be done later. You can recv and send messages from administrator here */
      FD_CLR(adm_src_fd, &f);
      n = recv(adm_src_fd, adm_buf, ADM_BUF_SIZE, 0);
      if(n < 0)
	  errx(1, "Error in recv! Exiting...");

      if(n == 0){
	/* Administrator has closed connection */
	FD_CLR(adm_src_fd, &temp_f);
	adm_src_fd = -1;
      }
    }
    else{
      /* This should never happen */
      printf("Warning! None of agent, administrator or new administrator fd is set.");
      printf("And select has returned a positive value\n");
    }
    
    f = temp_f;
    timeout = delete_expired_leases(&lease_list_head);
  }
}
