struct in6_addr {
  uint8_t  s6_addr[16];      /* IPv6 address */
};

/* The IPv6 address is stored in network byte order. */

struct in6_addr {
  union {
    uint8_t  _S6_u8[16];
    uint32_t _S6_u32[4];
    uint64_t _S6_u64[2];
  } _S6_un;
};

#define s6_addr _S6_un._S6_u8

/*********************************************************/

#define PF_INET6        AF_INET6

/* 
 * The PF_INET6 is used in the first argument to the socket() function
 * to indicate that an IPv6 socket is being created. 
 */

/*********************************************************/

struct sockaddr_in6 {
  sa_family_t     sin6_family;    /* AF_INET6 16 bit field */
  in_port_t       sin6_port;      /* transport layer port in n/w byte order */
  uint32_t        sin6_flowinfo;  /* IPv6 traffic class & flow info. */
  struct in6_addr sin6_addr;      /* IPv6 address */
  uint32_t        sin6_scope_id;  /* set of interfaces for a scope */
};

/*
 * The contents and interpretation of this member is specified in [1].  
 * The sin6_flowinfo field SHOULD be set to zero by an implementation prior 
 * to using the sockaddr_in6 structure by an application on receive 
 * operations.
 */

/* For 4.4 BSD the following is the struct. The above one was for 4.3 BSD */
struct sockaddr_in6 {
  uint8_t         sin6_len;       /* length of this struct: 8-bit */
  sa_family_t     sin6_family;    /* AF_INET6: 8-bit */
  in_port_t       sin6_port;      /* transport layer port # */
  uint32_t        sin6_flowinfo;  /* IPv6 flow information */
  struct in6_addr sin6_addr;      /* IPv6 address */
  uint32_t        sin6_scope_id;  /* set of interfaces for a scope */
};
/*********************************************************/


/******************* The socket functions ****************/

s = socket(PF_INET6, SOCK_STREAM, 0);  // IPv6/TCP socket
s = socket(PF_INET6, SOCK_DGRAM, 0);   // IPv6/UDP socket

/*********************************************************/
/*                    WILDCARD ADDRESS
 *                   ------------------
 * Since the IPv6 address type is a structure (struct in6_addr), a
 * symbolic constant can be used to initialize an IPv6 address variable,
 * but cannot be used in an assignment.  Therefore systems provide the
 * IPv6 wildcard address in two forms.
 * (1) extern const struct in6_addr in6addr_any;
         struct sockaddr_in6 sin6;
	 . . .
	 sin6.sin6_family = AF_INET6;
	 sin6.sin6_flowinfo = 0;
	 sin6.sin6_port = htons(23);
	 sin6.sin6_addr = in6addr_any;  // structure assignment 
	 . . .
         if (bind(s, (struct sockaddr *) &sin6, sizeof(sin6)) == -1)
	 . . .
 * (2) IN6ADDR_ANY_INIT: a symbolic constant ex.
         struct in6_addr anyaddr = IN6ADDR_ANY_INIT;
	 
	 But the following doesn't work
	 struct sockaddr_in6 sin6;
	 . . .
	 sin6.sin6_addr = IN6ADDR_ANY_INIT; // will NOT compile 

	              LOOPBACK ADDRESS
		     ------------------
    (1) extern const struct in6_addr in6addr_loopback;
          struct sockaddr_in6 sin6;
	  . . .
	  sin6.sin6_family = AF_INET6;
	  sin6.sin6_flowinfo = 0;
	  sin6.sin6_port = htons(23);
	  sin6.sin6_addr = in6addr_loopback;  
	  . . .
	  if (connect(s, (struct sockaddr *) &sin6, sizeof(sin6)) == -1)
	  . . .
     (2) IN6ADDR_LOOPBACK_INIT:
           struct in6_addr loopbackaddr = IN6ADDR_LOOPBACK_INIT;


	         Name-to-Index & Index-to-Name
                -------------------------------
     #include <net/if.h>
     unsigned int  if_nametoindex(const char *ifname);
     char  *if_indextoname(unsigned int ifindex, char *ifname);


             Return All Interface Names and Indexes
	    ----------------------------------------
     struct if_nameindex  *if_nameindex(void);

     struct if_nameindex {
       signed int   if_index;    // 1, 2, ... 
       char          *if_name;   //  null terminated name: "le0", ... 
     };
     
                        Free Memory 
		       -------------
    void  if_freenameindex(struct if_nameindex *ptr);

                       Socket Options
		      ----------------
    (1) setsockopt():
        int  hoplimit = 10;
        if (setsockopt(s, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
                     (char *) &hoplimit, sizeof(hoplimit)) == -1)
            perror("setsockopt IPV6_UNICAST_HOPS");

        The integer hop limit value (called x) is interpreted as follows:
	  x < -1:        return an error of EINVAL
	  x == -1:       use kernel default
	  0 <= x <= 255: use x
	  x >= 256:      return an error of EINVAL

    (2) getsockopt():
        int  hoplimit;
        size_t  len = sizeof(hoplimit);
	if (getsockopt(s, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
                     (char *) &hoplimit, &len) == -1)
            perror("getsockopt IPV6_UNICAST_HOPS");
        else
            printf("Using %d for hop limit.\n", hoplimit);


     Other options (other than "IPV6_UNICAST_HOPS") are
       - IPV6_MULTICAST_IF
         Set the interface to use for outgoing multicast packets.  The
         argument is the index of the interface to use.
         Argument type: unsigned int

       - IPV6_MULTICAST_HOPS
         Set the hop limit to use for outgoing multicast packets.
           x < -1:        return an error of EINVAL
           x == -1:       use kernel default
           0 <= x <= 255: use x
           x >= 256:      return an error of EINVAL

         If IPV6_MULTICAST_HOPS is not set, the default is 1 (same as IPv4)
         Argument type: int

       - IPV6_MULTICAST_LOOP
         If a multicast datagram is sent to a group to which the sending
         host itself belongs (on the outgoing interface), a copy of the
         datagram is looped back by the IP layer for local delivery if
         this option is set to 1.  If this option is set to 0 a copy
         is not looped back.  Other option values return an error of
         EINVAL. Default value is 1 i.e. loopback (same as Ipv4).
         Argument type: unsigned int

       - IPV6_JOIN_GROUP: error of EOPNOTSUPP if used with getsockopt().
         Join a multicast group on a specified local interface.  If the
         interface index is specified as 0, the kernel chooses the local
         interface.  For example, some kernels look up the multicast
         group in the normal IPv6 routing table and using the resulting
         interface.
         Argument type: struct ipv6_mreq

       - IPV6_LEAVE_GROUP: error of EOPNOTSUPP if used with getsockopt().
         Leave a multicast group on a specified interface.
         Argument type: struct ipv6_mreq

       Note that to receive multicast datagrams a process must join the
       multicast group and bind the UDP port to which datagrams will be
       sent.  Some processes also bind the multicast group address to the
       socket, in addition to the port, to prevent other datagrams destined
       to that same port from being delivered to the socket.


 */

/*
References

   [1]  Deering, S. and R. Hinden, "Internet Protocol, Version 6 (IPv6)
        Specification", RFC 2460, December 1998.

   [2]  Hinden, R. and S. Deering, "IP Version 6 Addressing
        Architecture", RFC 2373, July 1998.

   [3]  IEEE, "Protocol Independent Interfaces", IEEE Std 1003.1g, DRAFT
        6.6, March 1997.

   [4]  Stevens, W. and M. Thomas, "Advanced Sockets API for IPv6", RFC
        2292, February 1998.
*/
