# include <stdio.h>
# include <sys/types.h>
# include <stdlib.h>
# include <netinet/in.h>
# include <arpa/inet.h>

# define MAX_DHCP_PACKET  1218

#if defined (BYTE_ORDER) && (BYTE_ORDER == LITTLE_ENDIAN)
int global = 0;
#else
int global = 1;
#endif


typedef union{
  char             extensions[MAX_DHCP_PACKET];
  struct{
    uint32_t         preference                  :8  ;
    uint32_t         solicit_id                  :9  ;
    uint32_t         reserved                    :7  ;
    uint32_t         msg_type                    :8  ;
    
    struct in6_addr  client_link_local_address       ;
    struct in6_addr  relay_address                   ;
    struct in6_addr  server_address                  ;
    }_s_;
  
} test;

typedef struct{

  uint8_t          msg_type      ;
  uint8_t          reserved   : 6;
  uint8_t          p          : 1;
  uint8_t          c          : 1;
  union{

    struct{
      uint16_t         solicit_id : 9;
      uint16_t         prefix_len : 7;
    }_s_;

    uint16_t prefix_len_solicit_id;

  }_u_;

  
  struct in6_addr client_link_local_address;
  struct in6_addr relay_address;
  int test;
} s;

typedef struct{
  /*
  uint32_t         solicit_id                 : 9 ;
  uint32_t         prefix_len                 : 7 ;
  uint32_t         reserved                   : 6 ;
  uint32_t         p                          : 1 ;
  uint32_t         c                          : 1 ;
  uint32_t         msg_type                   : 8 ;
  */

  uint8_t          msg_type;
  uint8_t          c_p_reserved;
  uint16_t          prefix_len_solicit_id;
  struct in6_addr  client_link_local_address      ;
  struct in6_addr  relay_address                  ;

} solicit;

typedef struct{
  uint32_t         preference                 : 8  ;
  uint32_t         solicit_id                 : 9  ;
  uint32_t         reserved                   : 7  ;
  uint32_t         msg_type                   : 8  ;

  struct in6_addr  client_link_local_address       ;
  struct in6_addr  relay_address                   ;
  struct in6_addr  server_address                  ;
  /*
   * extensions needed
   */

} advertise;

typedef struct{
  uint32_t         transaction_id             : 16;
  uint32_t         reserved                   : 6 ;
  uint32_t         r                          : 1 ;
  uint32_t         c                          : 1 ;
  uint32_t         msg_type                   : 8 ;

  struct in6_addr  client_link_local_address      ;
  struct in6_addr  relay_address                  ;
  struct in6_addr  server_address                 ;
  /*
   * extensions needed
   */

} request;

typedef struct{

  uint32_t         transaction_id             : 16;
  uint32_t         status                     : 7 ;
  uint32_t         r                          : 1 ;
  uint32_t         msg_type                   : 8 ;

  struct in6_addr  client_link_local_address      ;
  struct in6_addr  relay_address                  ;
  /*
   * extensions needed
   */

} reply;

typedef struct{

  uint32_t         transaction_id             : 16;
  uint32_t         reserved                   : 7 ;
  uint32_t         r                          : 1 ;
  uint32_t         msg_type                   : 8 ;

  struct in6_addr  client_link_local_address      ;
  struct in6_addr  server_address                 ;
  struct in6_addr  x_address                      ;
  /*
   * extensions needed
   */

} release;

typedef struct{

  uint32_t         transaction_id             : 16;
  uint32_t         reserved                   : 8 ;
  uint32_t         msg_type                   : 8 ;

  struct in6_addr  server_address                 ;
  /*
   * extensions needed
   */

} reconfigure;

typedef struct{
  uint32_t         transaction_id             : 16;
  uint32_t         status                     : 7 ;
  uint32_t         r                          : 1 ;
  uint32_t         msg_type                   : 8 ;

  struct in6_addr  client_link_local_address      ;
  struct in6_addr  server_address                 ;

} reconfigure_reply;

typedef struct{

  uint32_t         transaction_id             : 16;
  uint32_t         reserved                   : 8 ;
  uint32_t         msg_type                   : 8 ;
  struct in6_addr  server_address                 ;
  /*
   * extensions needed
   */

} reconfigure_init ;

void print_binary(unsigned char c){
  unsigned char result[8];
  int i = 8;

  bzero(result, 9);

  while(i > 0){
    result[--i] = (unsigned char)(c % 2);
    c = c / 2;
  }
  result[8] = '\0';
  
  for(i = 0; i < 8; i++)
    printf("%u",result[i]);
  printf(" ");
}

# define set_prefix_len(s, x) (s.prefix_len_solicit_id = (s.prefix_len_solicit_id | ((x) << 9))         )
# define set_solicit_id(s, x) (s.prefix_len_solicit_id = (s.prefix_len_solicit_id & (127 << 9) ) + (x) )

main(){
  
  solicit           sol;
  advertise         adv;
  request           req;
  reply             rep;
  release           rel;
  reconfigure       rec;
  reconfigure_reply rec_rep;
  reconfigure_init  rec_init;
  test              t;
  s                 s1;
  uint8_t *res;
  char msg[128];

  unsigned char *b;
  int c = 0;
  int p = 1;
  int size, i;
  int prefix_len = 2;
  int solicit_id = 511;
  uint32_t *temp;
  uint16_t temp_16;


  res = (uint8_t *)&t;
  printf("res = %u global = %d\n",*res, global);

  size = 4;
  bzero((char *) &sol, sizeof(sol));
  bzero((char *) &adv, sizeof(adv));
  bzero((char *) &req, sizeof(req));
  bzero((char *) &rep, sizeof(rep));
  bzero((char *) &rel, sizeof(rel));
  bzero((char *) &rec, sizeof(rec));
  bzero((char *) &rec_rep, sizeof(rec_rep));
  bzero((char *) &rec_init, sizeof(rec_init));

  printf("sizeof sol = %u\n", sizeof sol);
  sol.msg_type                  = 1;
  sol.c_p_reserved              = (c << 7) | (p << 6);
  temp_16                       = htons((prefix_len << 9)  | (solicit_id));
  memcpy(&sol.prefix_len_solicit_id, &temp_16, 16);

  /*
  printf("\n\nsol:-> = %u:8 %u:1 %u:1 %u:6 %u:7 %u:9\n",
	 sol.msg_type,
	 sol.c,
	 sol.p,
	 sol.reserved,
	 sol.prefix_len,
	 sol.solicit_id);
  */

  for(i = 0, b = (unsigned char *) &sol; i < size; i++)
    print_binary(*(b++));

  memcpy(msg, &sol, sizeof(sol));
  res = (uint8_t *) msg;
  printf("msg_type = %u %u\n", *res, msg[0]);
  getchar();
  adv.msg_type = 2;
  adv.reserved   = 0;
  adv.solicit_id = 7+256;
  adv.preference = 9;
  printf("\n\nadv:-> = %u:8 %u:7 %u:9 %u:8\n",
	 adv.msg_type,
	 adv.reserved,
	 adv.solicit_id,
	 adv.preference);
  temp = (uint32_t *) &adv;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &adv; i < size; i++)
    print_binary(*(b++));

  req.msg_type   = 3;
  req.c          = 1;
  req.r          = 0;
  req.reserved   = 0;
  req.transaction_id = 1;
  printf("\n\nreq:-> = %u:8 %u:1 %u:1 %u:6 %u:16\n",
	 req.msg_type,
	 req.c,
	 req.r,
	 req.reserved,
	 req.transaction_id);
  temp = (uint32_t *) &req;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &req; i < size; i++)
    print_binary(*(b++));

  memcpy(msg, &req, sizeof(req));
  res = (uint8_t *) msg;
  printf("msg_type of 3 = %u %d\n", *res, msg[0]);
  getchar();

  rep.msg_type = 4;
  rep.r          = 1;
  rep.status     = 6;
  rep.transaction_id = 0;
  printf("\n\nrep:-> = %u:8 %u:1 %u:7 %u:16\n",
	 rep.msg_type,
	 rep.r,
	 rep.status,
	 rep.transaction_id);
  temp = (uint32_t *) &rep;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &rep; i < size; i++)
    print_binary(*(b++));

  rel.msg_type = 5;
  rel.r          = 0;
  rel.reserved   = 0;
  rel.transaction_id = 512;
  printf("\n\nrel:-> = %u:8 %u:1 %u:7 %u:16\n",
	 rel.msg_type,
	 rel.r,
	 rel.reserved,
	 rel.transaction_id);
  temp = (uint32_t *) &rel;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &rel; i < size; i++)
    print_binary(*(b++));

  rec.msg_type = 6;
  rec.reserved   = 1;
  rec.transaction_id = 1025;
  printf("\n\nrec:-> = %u:8 %u:8 %u:16\n",
	 rec.msg_type,
	 rec.reserved,
	 rec.transaction_id);
  temp = (uint32_t *) &rec;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &rec; i < size; i++)
    print_binary(*(b++));

  rec_rep.msg_type = 7;
  rec_rep.r          = 1;
  rec_rep.status     = 0;
  rec_rep.transaction_id = 0x00ff;
  printf("\n\nrec_rep:-> = %u:8 %u:1 %u:7 %u:16\n",
	 rec_rep.msg_type,
	 rec_rep.r,
	 rec_rep.status,
	 rec_rep.transaction_id);
  temp = (uint32_t *) &rec_rep;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &rec_rep; i < size; i++)
    print_binary(*(b++));

  rec_init.msg_type = 8;
  rec_init.reserved   = 127;
  rec_init.transaction_id = 0x0fff;
  printf("\n\nrec_init:-> = %u:8 %u:8 %u:16\n",
	 rec_init.msg_type,
	 rec_init.reserved,
	 rec_init.transaction_id);
  temp = (uint32_t *) &rec_init;
  *temp = htonl(*temp);
  for(i = 0, b = (unsigned char *) &rec_init; i < size; i++)
    print_binary(*(b++));

  printf("\n\n Press enter: ");/*
  sol.client_link_local_address.s6_addr32[0] = 0;
  sol.client_link_local_address.s6_addr32[1] = 0;
  sol.client_link_local_address.s6_addr32[2] = 0;
  sol.client_link_local_address.s6_addr32[3] = htonl(1);
  sol.relay_address.s6_addr32[0]             = 0;
  sol.relay_address.s6_addr32[1]             = 0;
  sol.relay_address.s6_addr32[2]             = 0;
  sol.relay_address.s6_addr32[3]             = 0;
			       */
  size = 4;

  printf("\nadv:-> = ");
  for(i = 0, b = (unsigned char *) &adv; i < size; i++)
    print_binary(*(b++));

  printf("\nbin:-> = ");
  for(i = 0, b = (unsigned char *) &sol; i < size; i++)
    print_binary(*(b++));

  printf("\nint = ");
  for(i = 0, b = (unsigned char *) &sol; i < size; i++)
    printf("%u ",*(b++));
  printf("\n");
  
  
  
  
  s1.msg_type                  = 16;
  s1.c                         = c;
  s1.p                         = p;
  s1.reserved                  = 0;
  s1._u_._s_.prefix_len = prefix_len;
  s1._u_._s_.solicit_id = solicit_id;

  s1._u_.prefix_len_solicit_id = htons(s1._u_.prefix_len_solicit_id);
  /*
  s1.client_link_local_address.s6_addr32[0] = 0;
  s1.client_link_local_address.s6_addr32[1] = 0;
  s1.client_link_local_address.s6_addr32[2] = 0;
  s1.client_link_local_address.s6_addr32[3] = 1;
  s1.relay_address.s6_addr32[0]             = 0;
  s1.relay_address.s6_addr32[1]             = 0;
  s1.relay_address.s6_addr32[2]             = 0;
  s1.relay_address.s6_addr32[3]             = 0;
  */
  s1.test                                   = htonl(1);

  size = sizeof(s1);

  printf("\nbin_s1 = ");
  for(i = 0, b = (unsigned char *) &s1; i < size; i++)
    print_binary(*(b++));

  printf("\nint = ");
  for(i = 0, b = (unsigned char *) &s1; i < size; i++)
    printf("%u ",*(b++));
  printf("\n");



  printf("solicit = %d adv = %d req = %d rep = %d rel = %d %d %d %d %d\n",
	 sizeof(solicit),
	 sizeof(advertise),
	 sizeof(request),
	 sizeof(reply),
	 sizeof(release),
	 sizeof(reconfigure),
	 sizeof(reconfigure_reply),
	 sizeof(reconfigure_init),
	 sizeof(test));
  
}
